# -*- coding: utf-8 -*-
"""
Created on Tue Feb  2 14:54:38 2021

@author: wanxiang.shen@u.nus.edu
"""


import numpy as np
import pandas as pd

from tqdm import tqdm
from copy import copy

from aggmap.utils.matrixopt import conv2
from sklearn.metrics import mean_squared_error, log_loss
from sklearn.preprocessing import StandardScaler



def islice(lst, n):
    return [lst[i:i + n] for i in range(0, len(lst), n)]


def GlobalIMP(clf, mp, X, Y, task_type = 'classification', 
              binary_task = False,
              sigmoidy = False, 
              apply_logrithm = False,
              apply_smoothing = False, 
              kernel_size = 5, 
              sigma = 1.6):
    '''
    Forward prop. Feature importance
    
    apply_scale_smothing: alpplying a smothing on the map
    
    '''

    if task_type == 'classification':
        f = log_loss
    else:
        f = mean_squared_error
        
    def sigmoid(x):
        return 1 / (1 + np.exp(-x))

    scaler = StandardScaler()
    df_grid = mp.df_grid_reshape
    backgroud = mp.transform_mpX_to_df(clf.X_).min().values #min value in the training set

    dfY = pd.DataFrame(Y)
    Y_true = Y
    Y_prob = clf._model.predict(X)
    N, W, H, C = X.shape
    T = len(df_grid)
    nX = 20 # 10 arrX to predict

    if (sigmoidy) & (task_type == 'classification'):
        Y_prob = sigmoid(Y_prob)
    

    final_res = {}
    for k, col in enumerate(dfY.columns):
        if (task_type == 'classification') & (binary_task):
            if k == 0:
                continue
        print('calculating feature importance for column %s ...' % col)
        results = []
        loss = f(Y_true[:, k].tolist(), Y_prob[:, k].tolist())
        
        tmp_X = []
        flag = 0
        for i in tqdm(range(T), ascii= True):
            ts = df_grid.iloc[i]
            y = ts.y
            x = ts.x
            
            ## step 1: make permutaions
            X1 = np.array(X)
            #x_min = X[:, y, x,:].min()
            vmin = backgroud[i]
            X1[:, y, x,:] = np.full(X1[:, y, x,:].shape, fill_value = vmin)
            tmp_X.append(X1)

            if (flag == nX) | (i == T-1):
                X2p = np.concatenate(tmp_X)
                ## step 2: make predictions
                Y_pred_prob = clf._model.predict(X2p) #predict ont by one is not efficiency
                if (sigmoidy) & (task_type == 'classification'):
                    Y_pred_prob = sigmoid(Y_pred_prob)    

                ## step 3: calculate changes
                for Y_pred in islice(Y_pred_prob, N):
                    mut_loss = f(Y_true[:, k].tolist(), Y_pred[:, k].tolist()) 
                    res =  mut_loss - loss # if res > 0, important, othervise, not important
                    results.append(res)

                flag = 0
                tmp_X = []
            flag += 1

        ## step 4:apply scaling or smothing 
        s = pd.DataFrame(results).values
        if apply_logrithm:
            s = np.log(s)
        smin = np.nanmin(s[s != -np.inf])
        smax = np.nanmax(s[s != np.inf])
        s = np.nan_to_num(s, nan=smin, posinf=smax, neginf=smin) #fillna with smin
        a = scaler.fit_transform(s)
        a = a.reshape(*mp._S.fmap_shape)
        if apply_smoothing:
            covda = conv2(a, kernel_size=kernel_size, sigma=sigma)
            results = covda.reshape(-1,).tolist()
        else:
            results = a.reshape(-1,).tolist()

        final_res.update({col:results})
        
    df = pd.DataFrame(final_res)
    df.columns = df.columns.astype(str)
    df.columns = 'col_' + df.columns + '_importance'
    df = df_grid.join(df)
    return df



def LocalIMP(clf, mp, X, Y, 
             task_type = 'classification', 
             binary_task = False,
             sigmoidy = False,  
             apply_logrithm = False, 
             apply_smoothing = False,
             kernel_size = 3, sigma = 1.2):
    '''
    Forward prop. Feature importance
    '''
    
    assert len(X) == 1, 'each for only one image!'
    
    if task_type == 'classification':
        f = log_loss
    else:
        f = mean_squared_error
        
    def sigmoid(x):
        return 1 / (1 + np.exp(-x))
    
    scaler = StandardScaler()
    df_grid = mp.df_grid_reshape
    backgroud = mp.transform_mpX_to_df(clf.X_).min().values #min value in the training set
    
    Y_true = Y
    Y_prob = clf._model.predict(X)
    N, W, H, C = X.shape

    if (sigmoidy) & (task_type == 'classification'):
        Y_prob = sigmoid(Y_prob)

    results = []
    loss = f(Y_true.ravel().tolist(),  Y_prob.ravel().tolist())
    
    all_X1 = []
    for i in tqdm(range(len(df_grid)), ascii= True):
        ts = df_grid.iloc[i]
        y = ts.y
        x = ts.x
        X1 = np.array(X)
        vmin = backgroud[i]
        X1[:, y, x,:] = np.full(X1[:, y, x,:].shape, fill_value = vmin)
        all_X1.append(X1)
        
    all_X = np.concatenate(all_X1)
    all_Y_pred_prob = clf._model.predict(all_X)

    for Y_pred_prob in all_Y_pred_prob:
        if (sigmoidy) & (task_type == 'classification'):
            Y_pred_prob = sigmoid(Y_pred_prob)
        mut_loss = f(Y_true.ravel().tolist(), Y_pred_prob.ravel().tolist()) 
        res =  mut_loss - loss # if res > 0, important, othervise, not important
        results.append(res)

    ## apply smothing and scalings
    s = pd.DataFrame(results).values
    if apply_logrithm:
        s = np.log(s)
    smin = np.nanmin(s[s != -np.inf])
    smax = np.nanmax(s[s != np.inf])
    s = np.nan_to_num(s, nan=smin, posinf=smax, neginf=smin) #fillna with smin
    a = scaler.fit_transform(s)
    a = a.reshape(*mp._S.fmap_shape)
    if apply_smoothing:
        covda = conv2(a, kernel_size=kernel_size, sigma=sigma)
        results = covda.reshape(-1,).tolist()
    else:
        results = a.reshape(-1,).tolist()

    df = pd.DataFrame(results, columns = ['imp'])
    #df.columns = df.columns + '_importance'
    df = df_grid.join(df)
    return df


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Aug 25 20:29:36 2019

@author: wanxiang.shen@u.nus.edu

main aggmap code

"""
from aggmap.utils.logtools import print_info, print_warn, print_error
from aggmap.utils.matrixopt import Scatter2Grid, Scatter2Array, smartpadding 
from aggmap.utils import vismap, summary, calculator
from aggmap.utils.gen_nwk import mp2newick


from sklearn.manifold import TSNE, MDS, Isomap, LocallyLinearEmbedding, SpectralEmbedding
        
from joblib import Parallel, delayed, load, dump
from scipy.spatial.distance import squareform, cdist, pdist
from scipy.cluster.hierarchy import fcluster, linkage, dendrogram
import matplotlib.pylab as plt
import seaborn as sns
from umap import UMAP
from tqdm import tqdm
from copy import copy, deepcopy
import pandas as pd
import numpy as np


class Base:
    
    def __init__(self):
        pass
        
    def _save(self, filename):
        return dump(self, filename)
        
    def _load(self, filename):
        return load(filename)

    
    def MinMaxScaleClip(self, x, xmin, xmax):
        scaled = (x - xmin) / ((xmax - xmin) + 1e-8)
        return scaled

    def StandardScaler(self, x, xmean, xstd):
        return (x-xmean) / (xstd + 1e-8) 
    
    
class Random_2DEmbedding:
    
    def __init__(self, random_state=123, n_components=2):
        self.random_state=random_state
        self.n_components = n_components

    def fit(self, X):
        M, N = X.shape
        np.random.seed(self.random_state)
        rho = np.sqrt(np.random.uniform(0, 1, N))
        phi = np.random.uniform(0, 4*np.pi, N)
        x = rho * np.cos(phi)
        y = rho * np.sin(phi)
        rd = pd.DataFrame([x,y]).T.sample(frac=1, random_state=123).reset_index(drop=True)
        self.embedding_ = rd.values
        return self
        
        
def _get_df_scatter(mp):
    xy = mp.embedded.embedding_
    colormaps = mp.colormaps
    df = pd.DataFrame(xy, columns = ['x', 'y'])
    bitsinfo = mp.bitsinfo.set_index('IDs')
    df = df.join(bitsinfo.loc[mp.flist].reset_index())
    df['colors'] = df['Subtypes'].map(colormaps)
    return df


def _get_df_grid(mp):

    p,q = mp._S.fmap_shape
    position = np.zeros(mp._S.fmap_shape, dtype='O').reshape(p*q,)
    position[mp._S.col_asses] = mp.flist
    position = position.reshape(p, q)
    if mp.fmap_shape != None:  
        m, n = mp.fmap_shape
        if (m > p) | (n > q):
            position = smartpadding(position, (m,n), constant_values=0)        
    M, N = position.shape
    
    x = []
    y = []
    for i in range(M):
        for j in range(N):
            x.append(j) #, position[j,i]
            y.append(i)
    v = position.reshape(M*N,)

    df = pd.DataFrame(list(zip(x,y, v)), columns = ['x', 'y', 'v'])

    bitsinfo = mp.bitsinfo
    subtypedict = bitsinfo.set_index('IDs')['Subtypes'].to_dict()
    subtypedict.update({0:'NaN'})
    df['Subtypes'] = df.v.map(subtypedict)
    df['colors'] = df['Subtypes'].map(mp.colormaps) 
    
    feature_list = df.v
    feature_names = []
    for i, j in feature_list.items():
        if j == 0:
            j = 'NaN-' + str(i)
        feature_names.append(j)
    df.v = feature_names
    return df


class AggMap(Base):
    
    def __init__(self, 
                 dfx,
                 metric = 'correlation',
                 by_scipy = False,
                 n_cpus = 16,
                 info_distance = None,
                ):
        
        """
        paramters
        -----------------
        dfx: pandas DataFrame
        metric: {'cosine', 'correlation', 'euclidean', 'jaccard', 'hamming', 'dice'}, default: 'correlation'
                measurement of feature distance

        info_distance: a vector-form distance vector of the feature points, shape should be: (n*(n-1)/2), where n is the number of the features, defalt: None
                        It can be useful when you have you own vector-form distance to pass
        by_scipy: bool, defalt: False.
                  calculate the distance by using the scipy pdist fuction.
                  It can bu useful when dfx.shape[1] > 20000, i.e., the number of features is very large
                  Using pdist will increase the speed to calculate the distance, but may result a lower precision
            
        n_cpus: int, default: 16
                number of cpu cores to use to calculate the distance.        
        """
        
        assert type(dfx) == pd.core.frame.DataFrame, 'input dfx must be pandas DataFrame!'
        super().__init__()
        self.metric = metric
        self.by_scipy = by_scipy
        self.isfit = False
        self.alist = dfx.columns.tolist()
        self.ftype = 'feature points'
        self.cluster_flag = False
        m,n = dfx.shape
        info_distance_length = int(n*(n-1)/2)
        assert len(dfx.columns.unique()) == n, 'the column names of dataframe must be unique !'        
        
        ## calculating distance
        if np.array(info_distance).any():
            assert len(info_distance) == info_distance_length, 'shape of info_distance must be (%s,)' % info_distance_length
            print_info('Skipping the distance calculation, using the customized vector-form distance...')
            self.info_distance = np.array(info_distance)
            self.metric = 'precomputed'
        else:
            print_info('Calculating distance ...')
            
            if self.by_scipy:
                D = pdist(dfx.T, metric=metric)
                D = np.nan_to_num(D,copy=False)
                self.info_distance = D.clip(0, np.inf)
            else:
                D = calculator.pairwise_distance(dfx.values, n_cpus=n_cpus, method=metric)
                D = np.nan_to_num(D,copy=False)
                D_ = squareform(D)
                self.info_distance = D_.clip(0, np.inf)
            
        ## statistic info
        S = summary.Summary(n_jobs = 10)
        res= []
        for i in tqdm(range(dfx.shape[1]), ascii=True):
            r = S._statistics_one(dfx.values, i)
            res.append(r)
        dfs = pd.DataFrame(res, index = self.alist)
        self.info_scale = dfs
        
        
        
    def _fit_embedding(self, 
                       dist_matrix,
                       emb_method = 'umap', 
                       n_components = 2,
                       random_state = 32,  
                       verbose = 2,
                       n_neighbors = 15,
                       min_dist = 0.1,
                       **kwargs):
        
        """
        parameters
        -----------------
        dist_matrix: distance matrix to fit
        emb_method: {'tsne', 'umap', 'mds'}, algorithm to embedd high-D to 2D
        kwargs: the extra parameters for the conresponding algorithm
        """
        

        if 'metric' in kwargs.keys():
            metric = kwargs.get('metric')
            kwargs.pop('metric')
            
        else:
            metric = 'precomputed'

        if emb_method == 'tsne':
            embedded = TSNE(n_components=n_components, 
                            random_state=random_state,
                            metric = metric,
                            verbose = verbose,
                            **kwargs)
            embedded = embedded.fit(dist_matrix)   
            
        elif emb_method == 'umap':
            embedded = UMAP(n_components = n_components, 
                            n_neighbors = n_neighbors,
                            min_dist = min_dist,
                            verbose = verbose,
                            random_state=random_state, 
                            metric = metric, **kwargs)
            embedded = embedded.fit(dist_matrix)   
            
        elif emb_method =='mds':
            if 'metric' in kwargs.keys():
                kwargs.pop('metric')
            if 'dissimilarity' in kwargs.keys():
                dissimilarity = kwargs.get('dissimilarity')
                kwargs.pop('dissimilarity')
            else:
                dissimilarity = 'precomputed'
                
            embedded = MDS(metric = True, 
                           n_components= n_components,
                           verbose = verbose,
                           dissimilarity = dissimilarity, 
                           random_state = random_state, **kwargs)
            embedded = embedded.fit(dist_matrix)          
        
        elif emb_method == 'random':
            embedded = Random_2DEmbedding(random_state=random_state, 
                                          n_components=n_components)
            embedded = embedded.fit(dist_matrix)   
            
        elif emb_method == 'isomap':
            embedded = Isomap(n_neighbors = n_neighbors,
                              n_components=n_components, 
                              metric = metric,
                              **kwargs)
            embedded = embedded.fit(dist_matrix)   
            
        elif emb_method == 'lle':
            embedded = LocallyLinearEmbedding(random_state=random_state, 
                                              n_neighbors = n_neighbors,
                                              n_components=n_components, 
                                              **kwargs)
            embedded = embedded.fit(dist_matrix)   
            
        elif emb_method == 'se':
            embedded = SpectralEmbedding(random_state=random_state, 
                                          n_neighbors = n_neighbors,
                                          n_components=n_components, 
                                          affinity = metric,
                                          **kwargs)
            affinity_matrix = np.exp(-dist_matrix**2)  #make more uniformly embedding  
            
            embedded = embedded.fit(affinity_matrix)    
    
        return embedded


    
    def fit(self, 
            feature_group_list = [],
            cluster_channels = 5,
            var_thr = -1, 
            split_channels = True, 
            fmap_shape = None, 
            emb_method = 'umap', 
            min_dist = 0.1, 
            n_neighbors = 15,
            verbose = 2, 
            random_state = 32,
            group_color_dict  = {},
            lnk_method = 'complete',
            **kwargs): 
        """
        parameters
        -----------------
        feature_group_list: list of the group name for each feature point
        cluster_channels: int, number of the channels(clusters) if feature_group_list is empty
        var_thr: float, defalt is -1, meaning that feature will be included only if the conresponding variance larger than this value. Since some of the feature has pretty low variances, we can remove them by increasing this threshold
        split_channels: bool, if True, outputs will split into various channels using the types of feature
        fmap_shape: None or tuple, size of molmap, if None, the size of feature map will be calculated automatically
        emb_method: {'tsne', 'umap', 'mds', 'isomap', 'random', 'lle', 'se'}, algorithm to embedd high-D to 2D
        group_color_dict: dict of the group colors, keys are the group names, values are the colors
        lnk_method: {'complete', 'average', 'single', 'weighted', 'centroid'}, linkage method
        kwargs: the extra parameters for the conresponding embedding method
        """
            
        if 'n_components' in kwargs.keys():
            kwargs.pop('n_components')
            
            
        ## embedding  into a 2d 
        assert emb_method in ['tsne', 'umap', 'mds', 'isomap', 'random', 'lle', 'se'], 'No Such Method Supported: %s' % emb_method

        self.var_thr = var_thr
        self.split_channels = split_channels
        self.fmap_shape = fmap_shape
        self.emb_method = emb_method
        self.lnk_method = lnk_method
        
        if fmap_shape != None:
            assert len(fmap_shape) == 2, "fmap_shape must be a tuple with two elements!"
        
        # flist and distance
        flist = self.info_scale[self.info_scale['var'] > self.var_thr].index.tolist()
        
        dfd = pd.DataFrame(squareform(self.info_distance),
                           index=self.alist,
                           columns=self.alist)
        dist_matrix = dfd.loc[flist][flist]
        self.flist = flist
        
        self.x_mean = self.info_scale['mean'].values
        self.x_std =  self.info_scale['std'].values
        
        self.x_min = self.info_scale['min'].values
        self.x_max = self.info_scale['max'].values
     
        #bitsinfo
        dfb = pd.DataFrame(self.flist, columns = ['IDs'])
        if feature_group_list != []:
            
            self.cluster_flag = False
            
            assert len(feature_group_list) == len(self.alist), "the length of the input group list is not equal to length of the feature list"
            self.cluster_channels = len(set(feature_group_list))
            self.feature_group_list = feature_group_list
            dfb['Subtypes'] = dfb['IDs'].map(pd.Series(feature_group_list, index = self.alist))
            
            if set(feature_group_list).issubset(set(group_color_dict.keys())):
                self.group_color_dict = group_color_dict
                dfb['colors'] = dfb['Subtypes'].map(group_color_dict)
            else:
                unique_types = dfb['Subtypes'].unique()
                color_list = sns.color_palette("hsv", len(unique_types)).as_hex()
                group_color_dict = dict(zip(unique_types, color_list))
                dfb['colors'] = dfb['Subtypes'].map(group_color_dict)
                self.group_color_dict = group_color_dict
        else:
            self.cluster_channels = cluster_channels
            print_info('applying hierarchical clustering to obtain group information ...')
            self.cluster_flag = True
            
            Z = linkage(squareform(dist_matrix.values),  lnk_method)
            labels = fcluster(Z, cluster_channels, criterion='maxclust')
            
            feature_group_list = ['cluster_%s' % str(i).zfill(2) for i in labels]
            dfb['Subtypes'] = feature_group_list
            dfb = dfb.sort_values('Subtypes')
            unique_types = dfb['Subtypes'].unique()
            
            if not set(unique_types).issubset(set(group_color_dict.keys())):
                color_list = sns.color_palette("hsv", len(unique_types)).as_hex()
                group_color_dict = dict(zip(unique_types, color_list))
            
            dfb['colors'] = dfb['Subtypes'].map(group_color_dict)
            self.group_color_dict = group_color_dict           
            self.Z = Z
            self.feature_group_list = feature_group_list
            

        self.bitsinfo = dfb
        colormaps = dfb.set_index('Subtypes')['colors'].to_dict()
        colormaps.update({'NaN': '#000000'})
        self.colormaps = colormaps
        self._S = Scatter2Grid()

        ## 2d embedding first
        embedded = self._fit_embedding(dist_matrix,
                                       emb_method = emb_method,
                                       n_neighbors = n_neighbors,
                                       random_state = random_state,
                                       min_dist = min_dist, 
                                       verbose = verbose,
                                       n_components = 2, **kwargs)
        
        self.embedded = embedded 
        
        df = pd.DataFrame(embedded.embedding_, index = self.flist,columns=['x', 'y'])
        typemap = self.bitsinfo.set_index('IDs')
        df = df.join(typemap)
        df['Channels'] = df['Subtypes']
        self.df_embedding = df
      
        ## linear assignment algorithm 
        print_info('Applying grid assignment of feature points, this may take several minutes(1~30 min)')
        self._S.fit(self.df_embedding, self.split_channels, channel_col = 'Channels')
        print_info('Finished')
        
        ## fit flag
        self.isfit = True
        if self.fmap_shape == None:
            self.fmap_shape = self._S.fmap_shape        
        else:
            m, n = self.fmap_shape
            p, q = self._S.fmap_shape
            assert (m >= p) & (n >=q), "fmap_shape's width must >= %s, height >= %s " % (p, q)

    
        self.df_scatter = _get_df_scatter(self)
        self.df_grid = _get_df_grid(self)
        self.df_grid_reshape = _get_df_grid(self)
        self.feature_names_reshape = self.df_grid.v.tolist()
        return self
        
    
    def refit_c(self, cluster_channels = 10, lnk_method = 'complete', group_color_dict = {}):
        """
        re-fit the aggmap object to update the number of channels
        parameters
        --------------------
        cluster_channels: int, number of the channels(clusters)
        group_color_dict: dict of the group colors, keys are the group names, values are the colors
        lnk_method: {'complete', 'average', 'single', 'weighted', 'centroid'}, linkage method
        """
        
        if not self.isfit:
            print_error('please fit first!')
            return
            

        self.split_channels = True
        self.lnk_method = lnk_method
        self.cluster_channels = cluster_channels
        
        dfd = pd.DataFrame(squareform(self.info_distance),
                           index=self.alist,
                           columns=self.alist)
        dist_matrix = dfd.loc[self.flist][self.flist]

        dfb = pd.DataFrame(self.flist, columns = ['IDs'])
        print_info('applying hierarchical clustering to obtain group information ...')
        self.cluster_flag = True

        Z = linkage(squareform(dist_matrix.values),  self.lnk_method)
        labels = fcluster(Z, self.cluster_channels, criterion='maxclust')

        feature_group_list = ['cluster_%s' % str(i).zfill(2) for i in labels]
        dfb['Subtypes'] = feature_group_list
        dfb = dfb.sort_values('Subtypes')
        unique_types = dfb['Subtypes'].unique()

        if not set(unique_types).issubset(set(group_color_dict.keys())):
            color_list = sns.color_palette("hsv", len(unique_types)).as_hex()
            group_color_dict = dict(zip(unique_types, color_list))

        dfb['colors'] = dfb['Subtypes'].map(group_color_dict)
        self.group_color_dict = group_color_dict           
        self.Z = Z
        self.feature_group_list = feature_group_list

        # update self.bitsinfo
        self.bitsinfo = dfb
        colormaps = dfb.set_index('Subtypes')['colors'].to_dict()
        colormaps.update({'NaN': '#000000'})
        self.colormaps = colormaps

        # update self.df_embedding
        df = pd.DataFrame(self.embedded.embedding_, index = self.flist,columns=['x', 'y'])
        typemap = self.bitsinfo.set_index('IDs')
        df = df.join(typemap)
        df['Channels'] = df['Subtypes']
        self.df_embedding = df

        ## linear assignment not performed, only refit the channel number 
        print_info('skipping grid assignment of feature points, fitting to target channel number')
        self._S.refit_c(self.df_embedding)
        print_info('Finished')

        if self.fmap_shape == None:
            self.fmap_shape = self._S.fmap_shape        
        else:
            m, n = self.fmap_shape
            p, q = self._S.fmap_shape
            assert (m >= p) & (n >=q), "fmap_shape's width must >= %s, height >= %s " % (p, q)


        self.df_scatter = _get_df_scatter(self)
        self.df_grid = _get_df_grid(self)
        self.df_grid_reshape = _get_df_grid(self)
        self.feature_names_reshape = self.df_grid.v.tolist()
        return self
    
    def transform_mpX_to_df(self, X):
        '''
        input 4D X, output 2D dataframe
        '''
        n, w,h, c = X.shape
        X_2D = X.sum(axis=-1).reshape(n, w*h)
        return pd.DataFrame(X_2D, columns = self.feature_names_reshape)    

    
    def transform(self, 
                  arr_1d, 
                  scale = True, 
                  scale_method = 'minmax',
                  fillnan = 0):
    
    
        """
        parameters
        --------------------
        arr_1d: 1d numpy array feature points
        scale: bool, if True, we will apply MinMax scaling by the precomputed values
        scale_method: {'minmax', 'standard'}
        fillnan: fill nan value, default: 0
        """
        
        if not self.isfit:
            print_error('please fit first!')
            return

        if scale:
            if scale_method == 'standard':
                arr_1d = self.StandardScaler(arr_1d, self.x_mean, self.x_std)
            else:
                arr_1d = self.MinMaxScaleClip(arr_1d, self.x_min, self.x_max)
        
        df = pd.DataFrame(arr_1d).T
        df.columns = self.alist

        df = df[self.flist]
        vector_1d = df.values[0] #shape = (N, )
        fmap = self._S.transform(vector_1d)  
        p, q, c = fmap.shape
        
        if self.fmap_shape != None:        
            m, n = self.fmap_shape
            if (m > p) | (n > q):
                fps = []
                for i in range(c):
                    fp = smartpadding(fmap[:,:,i], self.fmap_shape)
                    fps.append(fp)
                fmap = np.stack(fps, axis=-1)        

        return np.nan_to_num(fmap, nan = fillnan)   
    
    

    
    def batch_transform(self, 
                        array_2d, 
                        scale = True, 
                        scale_method = 'minmax',
                        n_jobs=4,
                        fillnan = 0):
    
        """
        parameters
        --------------------
        array_2d: 2D numpy array feature points, M(samples) x N(feature ponits)
        scale: bool, if True, we will apply MinMax scaling by the precomputed values
        scale_method: {'minmax', 'standard'}
        n_jobs: number of parallel
        fillnan: fill nan value, default: 0
        """
        
        if not self.isfit:
            print_error('please fit first!')
            return
        
        assert type(array_2d) == np.ndarray, 'input must be numpy ndarray!' 
        assert array_2d.ndim == 2, 'input must be 2-D  numpy array!' 
        
        P = Parallel(n_jobs=n_jobs)
        res = P(delayed(self.transform)(arr_1d, 
                                        scale,
                                        scale_method,
                                        fillnan) for arr_1d in tqdm(array_2d, ascii=True)) 
        X = np.stack(res) 
        
        return X
    
    
    def plot_scatter(self, htmlpath='./', htmlname=None, radius = 2, enabled_data_labels = False):
        """radius: the size of the scatter, must be int"""
        H_scatter = vismap.plot_scatter(self,  
                                        htmlpath=htmlpath, 
                                        htmlname=htmlname,
                                        radius = radius,
                                        enabled_data_labels = enabled_data_labels)
        return H_scatter   
        
        
    def plot_grid(self, htmlpath='./', htmlname=None, enabled_data_labels = False):
        
        H_grid = vismap.plot_grid(self,
                                  htmlpath=htmlpath, 
                                  htmlname=htmlname,
                                  enabled_data_labels = enabled_data_labels)
        return H_grid       
        
        
        
    def plot_tree(self, figsize=(16,8), add_leaf_labels = True, leaf_font_size = 18, leaf_rotation = 90):

        fig = plt.figure(figsize=figsize)
        
        if self.cluster_flag:
            
            Z = self.Z
            

            D_leaf_colors = self.bitsinfo['colors'].to_dict() 
            link_cols = {}
            for i, i12 in enumerate(Z[:,:2].astype(int)):
                c1, c2 = (link_cols[x] if x > len(Z) else D_leaf_colors[x] for x in i12)
                link_cols[i+1+len(Z)] = c1
            
            if add_leaf_labels:
                labels = self.alist
            else:
                labels = None
            P =dendrogram(Z, labels = labels, 
                          leaf_rotation = leaf_rotation, 
                          leaf_font_size = leaf_font_size, 
                          link_color_func=lambda x: link_cols[x])
        
        return fig
        
        
    def to_nwk_tree(self, treefile = 'mytree', leaf_names = None):
        '''
        convert mp object to newick tree and the data file to submit to itol sever
        '''
        return mp2newick(self, treefile = treefile, leaf_names=leaf_names)
        
        
    def copy(self):
        return deepcopy(self)
        
        
    def load(self, filename):
        return self._load(filename)
    
    
    def save(self, filename):
        return self._save(filename)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 17 16:54:12 2019

@author: wanxiang.shen@u.nus.edu

@usecase: statistic features' distribution
"""



import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tqdm import tqdm


from aggmap.utils.multiproc import MultiProcessUnorderedBarRun


class Summary:

    def __init__(self, n_jobs=1):
        
        '''
        n_jobs: number of paralleles
        '''
        self.n_jobs = n_jobs
        
        
    def _statistics_one(self, data, i):
        onefeat = data[:,i]
        
        
        onefeat = onefeat[~np.isnan(onefeat)]
        onefeat = onefeat[~np.isinf(onefeat)]
        
        s = pd.Series(onefeat)

        
        if len(s) != 0:
            maxv = s.max()
            minv = s.min()
            # using 0.8*(1-0.8) as a threshold to exclude feature
            var = s.var() 
            std = s.std()
            mean = s.mean()
            med = s.median()
            #skewness gt 0.75 will be log transformed
            skewness = s.skew()
            mode = s.mode().iloc[0]
        else:
            maxv = np.nan
            minv = np.nan
            var = np.nan
            std = np.nan
            mean = np.nan
            med = np.nan
            skewness = np.nan
            mode = np.nan
        
        del onefeat
        
        return {'index':i, 'max':maxv, 'mean':mean, 'min':minv, 'median':med, 
                'mode':mode, 'skewness':skewness, 'std':std, 'var': var}
    
    
    def fit(self, data, backend = 'threading', **kwargs):
        '''
        Parameters
        ----------
        data: np.memmap or np.array
        '''

        P = Parallel(n_jobs=self.n_jobs, backend = backend, **kwargs)
        res = P(delayed(self._statistics_one)(data,i) for i in tqdm(range(data.shape[1]), ascii=True))
        return pd.DataFrame(res)
    

def _func(i):
    S = Summary()
    return S._statistics_one(DATA, i)
    
    
def Summary2(data, n_jobs):
    global DATA, _func
    DATA = data
    res = MultiProcessUnorderedBarRun(_func, list(range(data.shape[1])), n_jobs)
    df = pd.DataFrame(res)
    dfres = df.sort_values('index').set_index('index')
    return dfres



# -*- coding: utf-8 -*-
"""
Created on Sun Aug 16 17:10:53 2020

@author: wanxiang.shen@u.nus.edu
"""

import warnings, os
warnings.filterwarnings("ignore")


import numpy as np
import pandas as pd
import tensorflow as tf
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.utils.multiclass import unique_labels
from sklearn.metrics import get_scorer, SCORERS

from aggmap import aggmodel
from aggmap.aggmodel.explain_dev import GlobalIMP, LocalIMP
from aggmap.aggmodel.explainer import shapley_explainer, simply_explainer

from joblib import dump, load
from  copy import copy
from tensorflow.keras.models import load_model as load_tf_model

import gc
import tensorflow.keras.backend as K
import tensorflow as tf


def clean(clf): 
    del clf._model
    del clf._performance
    del clf
    gc.collect()
    K.clear_session()
    tf.compat.v1.reset_default_graph() # TF graph isn't same as Keras graph
    
    
def save_model(model, model_path):
    if not os.path.exists(model_path):
        os.makedirs(model_path)
    print('saving model to %s' % model_path)
    model_new = copy(model)
    model_new._model.save(os.path.join(model_path, 'inner_model.h5'))
    model_new._model = None
    model_new._performance = None
    res = dump(model_new,  os.path.join(model_path, 'outer_model.est'))
    return res
    
def load_model(model_path, gpuid=None):
    '''
    gpuid: load model to specific gpu: {None, 0, 1, 2, 3,..}
    '''
    model = load(os.path.join(model_path, 'outer_model.est'))
    if gpuid==None:
        gpuid = model.gpuid
    else:
        gpuid = str(gpuid)
    os.environ["CUDA_VISIBLE_DEVICES"]= gpuid
    model.gpuid = gpuid
    model._model = load_tf_model(os.path.join(model_path, 'inner_model.h5'))
    return model


class RegressionEstimator(BaseEstimator, RegressorMixin):
    """ An AggMap CNN Regression estimator (each sample belongs to only one class) 
    Parameters
    ----------
    epochs : int, default = 200
        A parameter used for training epochs. 
    conv1_kernel_size: int, default = 13
        A parameter used for the kernel size of first covolutional layers
    dense_layers: list, default = [128]
        A parameter used for the dense layers.  
    batch_size: int, default: 128
        A parameter used for the batch size.
    lr: float, default: 1e-4
        A parameter used for the learning rate.
    loss:string or function, default: 'mse'
        A parameter used for the loss function
    batch_norm: bool, default: False
        batch normalization after convolution layers.
    n_inception: int, default:2
        Number of the inception layers.
    dense_avf: str, default is 'relu'
        activation fuction in the dense layers.
    dropout: float, default: 0
        A parameter used for the dropout of the dense layers.
    monitor: str, default: 'val_loss'
        {'val_loss', 'val_r2'}, a monitor for model selection
    metric: str, default: 'r2'
        {'r2', 'rmse'},  a matric parameter
    patience: int, default = 10000, 
        A parameter used for early stopping
    gpuid: int, default=0,
        A parameter used for specific gpu card
    verbose: int, default = 0
        if positive, then the log infomation of AggMapNet will be print
        if negative, then the log infomation of orignal model will be print
    random_state, int, default: 32

    Examples
    --------
    >>> from aggmap import AggModel
    >>> clf = AggModel.RegressionEstimator()

    """
    
    def __init__(self, 
                 epochs = 200,  
                 conv1_kernel_size = 13,
                 dense_layers = [128],  
                 dense_avf = 'relu',
                 batch_size = 128,  
                 lr = 1e-4, 
                 loss = 'mse',
                 batch_norm = False,
                 n_inception = 2,
                 dropout = 0.0,
                 monitor = 'val_loss', 
                 metric = 'r2',
                 patience = 10000,
                 verbose = 0, 
                 random_state = 32,
                 gpuid = 0,
                ):
        
        
        self.epochs = epochs
        self.dense_layers = dense_layers
        self.conv1_kernel_size = conv1_kernel_size
        self.dense_avf = dense_avf
        self.batch_size = batch_size
        self.lr = lr
        self.loss = loss
        self.batch_norm = batch_norm
        self.n_inception = n_inception
        self.dropout = dropout
        self.monitor = monitor
        self.metric = metric
        self.patience = patience
        
        self.gpuid = str(gpuid)
        os.environ["CUDA_VISIBLE_DEVICES"]= self.gpuid
        
        self.verbose = verbose
        self.random_state = random_state
        self.is_fit = False        
        self.name = "AggMap Regression Estimator"
    
        #print(self.get_params())
        
        self.history = {}
        self.history_model = {}
        
        if self.verbose > 0:
            self.verbose1 = self.verbose
            self.verbose2 = 0
        elif self.verbose ==0:
            self.verbose1 = 0
            self.verbose2 = 0
        elif self.verbose < 0:
            self.verbose1 = 0
            self.verbose2 = abs(self.verbose)    
        print(self)
        
    def get_params(self, deep=True):

        model_paras =  {"epochs": self.epochs, 
                        "lr":self.lr, 
                        "loss":self.loss,
                        "conv1_kernel_size": self.conv1_kernel_size,
                        "dense_layers": self.dense_layers, 
                        "dense_avf":self.dense_avf, 
                        "batch_size":self.batch_size, 
                        "dropout":self.dropout,
                        "batch_norm":self.batch_norm,
                        "n_inception":self.n_inception,
                        "monitor": self.monitor,
                        "metric":self.metric,
                        "patience":self.patience,
                        "random_state":self.random_state,
                        "verbose":self.verbose,
                        "name":self.name,
                        "gpuid": self.gpuid,
                       }

        return model_paras
    
    
    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self

        

    def fit(self, X, y,  X_valid = None, y_valid = None):

        # Check that X and y have correct shape
        
        if  X.ndim != 4:
            raise ValueError("Found array X with dim %d. %s expected == 4." % (X.ndim, self.name))
            
        if  y.ndim != 2:
            raise ValueError("Found array y with dim %d. %s expected == 2." % (y.ndim, self.name))    

        self.X_ = X
        self.y_ = y
        
        if (X_valid is None) | (y_valid is None):
            X_valid = X
            y_valid = y
        
        np.random.seed(self.random_state)
        tf.compat.v1.set_random_seed(self.random_state)
        

        model = aggmodel.net._AggMapNet(X.shape[1:],
                                       n_outputs = y.shape[-1], 
                                       conv1_kernel_size = self.conv1_kernel_size,
                                       batch_norm = self.batch_norm,
                                       n_inception = self.n_inception,
                                       dense_layers = self.dense_layers, 
                                       dense_avf = self.dense_avf, 
                                       dropout = self.dropout,
                                       last_avf = 'linear')  


        opt = tf.keras.optimizers.Adam(lr=self.lr, beta_1=0.9, beta_2=0.999, epsilon=1e-08, decay=0.0) #
        model.compile(optimizer = opt, loss = self.loss)        
        performance = aggmodel.cbks.Reg_EarlyStoppingAndPerformance((X, y), 
                                                                    (X_valid, y_valid), 
                                                                    patience = self.patience, 
                                                                    criteria = self.monitor,
                                                                    verbose = self.verbose1,)

        history = model.fit(X, y, 
                            batch_size=self.batch_size, 
                            epochs= self.epochs, verbose= self.verbose2, shuffle = True, 
                            validation_data = (X_valid, y_valid), 
                            callbacks=[performance]) 

        self._model = model
        self._performance = performance
        self.history = self._performance.history
        self.history_model = history.history
        self.is_fit = True
        # Return the classifier
        return self


    
    def predict(self, X):
        """
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features_w, n_features_h, n_features_c)
            Vector to be scored, where `n_samples` is the number of samples and

        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Returns the predicted values
        """
        y_pred = self._model.predict(X)
        return y_pred
    


    def score(self, X, y, scoring = 'r2', sample_weight=None):
        """Returns the score using the `scoring` option on the given
        test data and labels.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Test samples.
        y : array-like of shape (n_samples,)
            True labels for X.
        scoring: str, please refer to: https://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter
        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights.
        Returns
        -------
        score : float
            Score of self.predict(X) wrt. y.
        """
        assert scoring in SCORERS.keys(), 'scoring is not in %s' % SCORERS.keys()
        scoring = get_scorer(scoring)

        return scoring(self, X, y, sample_weight=sample_weight)  
    
    
    def plot_model(self, to_file='model.png', 
                   show_shapes=True, 
                   show_layer_names=True, 
                   rankdir='TB', 
                   expand_nested=False, 
                   dpi=96):
        if self.is_fit:
            tf.keras.utils.plot_model(self._model, 
                       to_file=to_file, 
                       show_shapes=show_shapes, 
                       show_layer_names=show_layer_names, 
                       rankdir=rankdir, 
                       expand_nested=expand_nested, 
                       dpi=dpi)
        else:
            print('Please fit first!')
    
    

    def save_model(self, model_path):
        return save_model(self, model_path)

    
    def load_model(self, model_path, gpuid=None):
        return load_model(model_path, gpuid=gpuid)

    
    def explain_model(self, mp, X, y, 
                      explain_format = 'global',
                      apply_logrithm = False, 
                      apply_smoothing = False, 
                      kernel_size = 3, sigma = 1.2):
        '''
        Feature importance calculation
        Parameters
        --------------
        mp: aggmap object
        X: trianing or test set X arrays
        y: trianing or test set y arrays
        explain_format: {'local', 'global'}, default: 'global'
            local or global feature importance, if local, then X must be one sample
        apply_logrithm: {True, False}, default: False
            whether apply a logarithm transformation on the importance values
        apply_smoothing: {True, False}, default: False
            whether apply a smoothing transformation on the importance values
        kernel_size: odd number, the kernel size to perform the smoothing
        sigma: float, sigma for gaussian smoothing
        
        Returns
        ------------        
        DataFrame of feature importance
        '''
        
        if explain_format == 'global':
            explain_func = GlobalIMP
        else:
            explain_func = LocalIMP
        
        dfe = explain_func(self, mp, X, y, 
                        task_type = 'regression', 
                        sigmoidy = False,  
                        apply_logrithm = apply_logrithm, 
                        apply_smoothing = apply_smoothing,
                        kernel_size = kernel_size, sigma = sigma)
        return dfe    
    
    @property
    def clean(self):
        clean(self)
        
    
class MultiClassEstimator(BaseEstimator, ClassifierMixin):

    """ An AggMap CNN MultiClass estimator (each sample belongs to only one class) 
    Parameters
    ----------
    epochs : int, default = 200
        A parameter used for training epochs. 
    conv1_kernel_size: int, default = 13
        A parameter used for the kernel size of first covolutional layers
    dense_layers: list, default = [128]
        A parameter used for the dense layers.  
    batch_size: int, default: 128
        A parameter used for the batch size.
    lr: float, default: 1e-4
        A parameter used for the learning rate.
    loss: string or function, default: 'categorical_crossentropy'
        A parameter used for the loss function
    batch_norm: bool, default: False
        batch normalization after convolution layers.
    n_inception: int, default:2
        Number of the inception layers.
    dense_avf: str, default is 'relu'
        activation fuction in the dense layers.
    dropout: float, default: 0
        A parameter used for the dropout of the dense layers.
    monitor: str, default: 'val_loss'
        {'val_loss', 'val_metric'}, a monitor for model selection
    metric: str, default: 'ACC'
        {'ROC', 'ACC', 'PRC'},  a matric parameter
    patience: int, default = 10000, 
        A parameter used for early stopping
    gpuid: int, default=0,
        A parameter used for specific gpu card
    verbose: int, default = 0
        if positive, then the log infomation of AggMapNet will be print
        if negative, then the log infomation of orignal model will be print
    random_state, int, default: 32


    Examples
    --------
    >>> from aggmap import AggModel
    >>> clf = AggModel.MultiClassEstimator()
    """
    
    
    def __init__(self, 
                 epochs = 200,  
                 conv1_kernel_size = 13,
                 dense_layers = [128],  
                 dense_avf = 'relu',
                 batch_size = 128,  
                 lr = 1e-4, 
                 loss = 'categorical_crossentropy', 
                 batch_norm = False,
                 n_inception = 2,                 
                 dropout = 0.0,
                 monitor = 'val_loss', 
                 metric = 'ACC',
                 patience = 10000,
                 verbose = 0, 
                 last_avf = 'softmax', 
                 random_state = 32,
                 gpuid=0,
                ):
        
        
        self.epochs = epochs
        self.dense_layers = dense_layers
        self.conv1_kernel_size = conv1_kernel_size
        self.dense_avf = dense_avf
        self.batch_size = batch_size
        self.lr = lr
        self.loss = loss
        self.last_avf = last_avf
        
        self.batch_norm = batch_norm
        self.n_inception = n_inception      
        self.dropout = dropout
        
        self.monitor = monitor
        self.metric = metric
        self.patience = patience
        
        self.gpuid = str(gpuid)
        os.environ["CUDA_VISIBLE_DEVICES"]= self.gpuid
        
        self.verbose = verbose
        self.random_state = random_state
        
        self.name = "AggMap MultiClass Estimator"
        self.is_fit = False        
        #print(self.get_params())
        self.history = {}        
        self.history_model = {}

        if self.verbose > 0:
            self.verbose1 = self.verbose
            self.verbose2 = 0
        elif self.verbose ==0:
            self.verbose1 = 0
            self.verbose2 = 0
        elif self.verbose < 0:
            self.verbose1 = 0
            self.verbose2 = abs(self.verbose)
            
        print(self)
        
    def get_params(self, deep=True):

        model_paras =  {"epochs": self.epochs, 
                        "lr":self.lr, 
                        "loss":self.loss,
                        "conv1_kernel_size": self.conv1_kernel_size,
                        "dense_layers": self.dense_layers, 
                        "dense_avf":self.dense_avf, 
                        "last_avf":self.last_avf,
                        "batch_size":self.batch_size, 
                        "dropout":self.dropout,
                        "batch_norm":self.batch_norm,
                        "n_inception":self.n_inception,                        
                        "monitor": self.monitor,
                        "metric":self.metric,
                        "patience":self.patience,
                        "random_state":self.random_state,
                        "verbose":self.verbose,
                        "name":self.name,
                        "gpuid": self.gpuid,
                       }

        return model_paras
    
    
    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self

        

    def fit(self, X, y,  
            X_valid = None,
            y_valid = None, 
            class_weight = None,
           ):

        # Check that X and y have correct shape
        
        if  X.ndim != 4:
            raise ValueError("Found array X with dim %d. %s expected == 4." % (X.ndim, self.name))
            
        if  y.ndim != 2:
            raise ValueError("Found array y with dim %d. %s expected == 2." % (y.ndim, self.name))    
    
        # Store the classes seen during fit
        self.classes_ = unique_labels(y)

        self.X_ = X
        self.y_ = y
        
        if (X_valid is None) | (y_valid is None):
            
            X_valid = X
            y_valid = y
        
        np.random.seed(self.random_state)
        tf.compat.v1.set_random_seed(self.random_state)


        model = aggmodel.net._AggMapNet(X.shape[1:],
                                        n_outputs = y.shape[-1], 
                                        conv1_kernel_size = self.conv1_kernel_size,
                                        batch_norm = self.batch_norm,
                                        n_inception = self.n_inception,
                                        dense_layers = self.dense_layers, 
                                        dense_avf = self.dense_avf, 
                                        dropout = self.dropout,
                                        last_avf = self.last_avf)


        opt = tf.keras.optimizers.Adam(lr=self.lr, beta_1=0.9, beta_2=0.999, epsilon=1e-08, decay=0.0) #
        model.compile(optimizer = opt, loss = self.loss, metrics = ['accuracy'])
            
        performance = aggmodel.cbks.CLA_EarlyStoppingAndPerformance((X, y), 
                                                                    (X_valid, y_valid), 
                                                                    patience = self.patience, 
                                                                    criteria = self.monitor,
                                                                    metric = self.metric,  
                                                                    last_avf= self.last_avf,
                                                                    verbose = self.verbose1,)

        history = model.fit(X, y, 
                  batch_size=self.batch_size, 
                  epochs= self.epochs, verbose= self.verbose2, shuffle = True, 
                  validation_data = (X_valid, y_valid), class_weight = class_weight, 
                  callbacks=[performance]) 

        self._model = model
        self._performance = performance
        self.history = self._performance.history
        self.history_model = history.history
        self.is_fit = True        
        # Return the classifier
        return self



    def predict_proba(self, X):
        """
        Probability estimates.
        The returned estimates for all classes are ordered by the
        label of classes.
        For a multi_class problem, if multi_class is set to be "multinomial"
        the softmax function is used to find the predicted probability of
        each class.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Vector to be scored, where `n_samples` is the number of samples and
            `n_features` is the number of features.
        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Returns the probability of the sample for each class in the model,
            where classes are ordered as they are in ``self.classes_``.
        """
        # Check is fit had been called
        check_is_fitted(self)
        
        # Input validation
        if  X.ndim != 4:
            raise ValueError("Found array X with dim %d. %s expected == 4." % (X.ndim, self.name))
        y_prob = self._model.predict(X)
        return y_prob
    

    def predict(self, X):
        probs = self.predict_proba(X)
        y_pred = np.argmax(probs, axis=1)
        return y_pred
    
    

    def score(self, X, y, scoring = 'accuracy', sample_weight=None):
        """Returns the score using the `scoring` option on the given
        test data and labels.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Test samples.
        y : array-like of shape (n_samples,)
            True labels for X.
        scoring: str, please refer to: https://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter
        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights.
        Returns
        -------
        score : float
            Score of self.predict(X) wrt. y.
        """
        assert scoring in SCORERS.keys(), 'scoring is not in %s' % SCORERS.keys()
        scoring = get_scorer(scoring)

        return scoring(self, X, y, sample_weight=sample_weight)
    
    
    def plot_model(self, to_file='model.png', 
                   show_shapes=True, 
                   show_layer_names=True, 
                   rankdir='TB', 
                   expand_nested=False, 
                   dpi=96):
        if self.is_fit:
            tf.keras.utils.plot_model(self._model, 
                       to_file=to_file, 
                       show_shapes=show_shapes, 
                       show_layer_names=show_layer_names, 
                       rankdir=rankdir, 
                       expand_nested=expand_nested, 
                       dpi=dpi)
        else:
            print('Please fit first!')    
    
    
    def save_model(self, model_path):
        return save_model(self, model_path)

    
    def load_model(self, model_path, gpuid=None):
        return load_model(model_path, gpuid=gpuid)

    
    def explain_model(self, mp, X, y, 
                      binary_task = False,
                      explain_format = 'global',
                      apply_logrithm = False, 
                      apply_smoothing = False, 
                      kernel_size = 3, sigma = 1.2):
        
        '''
        Feature importance calculation
        Parameters
        --------------
        mp: aggmap object
        X: trianing or test set X arrays
        y: trianing or test set y arrays
        binary_task: {True, False}
            whether the task is binary, if True, the feature importance will be calculated for one class only
        explain_format: {'local', 'global'}, default: 'global'
            local or global feature importance, if local, then X must be one sample
        apply_logrithm: {True, False}, default: False
            whether apply a logarithm transformation on the importance values
        apply_smoothing: {True, False}, default: False
            whether apply a smoothing transformation on the importance values
        kernel_size: odd number, the kernel size to perform the smoothing
        sigma: float, sigma for gaussian smoothing
        
        Returns
        ------------        
        DataFrame of feature importance
        '''
        if explain_format == 'global':
            explain_func = GlobalIMP
        else:
            explain_func = LocalIMP
        
        dfe = explain_func(self, mp, X, y, 
                           binary_task = binary_task,
                           task_type = 'classification',                            
                           sigmoidy = False,  
                           apply_logrithm = apply_logrithm, 
                           apply_smoothing = apply_smoothing,
                           kernel_size = kernel_size, sigma = sigma)
        return dfe
    
    @property
    def clean(self):
        clean(self)
        
        
class MultiLabelEstimator(BaseEstimator, ClassifierMixin):


    """ An AggMap CNN MultiLabel estimator (each sample belongs to only one class) 
    Parameters
    ----------
    epochs : int, default = 200
        A parameter used for training epochs. 
    conv1_kernel_size: int, default = 13
        A parameter used for the kernel size of first covolutional layers
    dense_layers: list, default = [128]
        A parameter used for the dense layers.  
    batch_size: int, default: 128
        A parameter used for the batch size.
    lr: float, default: 1e-4
        A parameter used for the learning rate.
    loss: string or function, default: tf.nn.sigmoid_cross_entropy_with_logits
        A parameter used for the loss function
    batch_norm: bool, default: False
        batch normalization after convolution layers.
    n_inception: int, default:2
        Number of the inception layers.
    dense_avf: str, default is 'relu'
        activation fuction in the dense layers.
    dropout: float, default: 0
        A parameter used for the dropout of the dense layers, such as 0.1, 0.3, 0.5.
    monitor: str, default: 'val_loss'
        {'val_loss', 'val_metric'}, a monitor for model selection
    metric: str, default: 'ROC'
        {'ROC', 'ACC', 'PRC'},  a matric parameter
    patience: int, default = 10000, 
        A parameter used for early stopping
    gpuid: int, default=0,
        A parameter used for specific gpu card
    verbose: int, default = 0
        if positive, then the log infomation of AggMapNet will be print
        if negative, then the log infomation of orignal model will be print
    random_state, int, default: 32
    name: str 

    Examples
    --------
    >>> from aggmap import AggModel
    >>> clf = AggModel.MultiLabelEstimator()
    """
    
    def __init__(self, 
                 epochs = 200,  
                 conv1_kernel_size = 13,
                 dense_layers = [128],  
                 dense_avf = 'relu',
                 batch_size = 128,  
                 lr = 1e-4, 
                 loss = tf.nn.sigmoid_cross_entropy_with_logits, 
                 batch_norm = False,
                 n_inception = 2,                     
                 dropout = 0.0,                 
                 monitor = 'val_loss', 
                 metric = 'ROC',
                 patience = 10000,
                 verbose = 0, 
                 random_state = 32,
                 gpuid = 0,
                ):
        
        
        self.epochs = epochs
        self.dense_layers = dense_layers
        self.conv1_kernel_size = conv1_kernel_size
        self.dense_avf = dense_avf
        self.batch_size = batch_size
        self.lr = lr
        self.loss = loss
        self.batch_norm = batch_norm
        self.n_inception = n_inception
        self.dropout = dropout
        self.monitor = monitor
        self.metric = metric
        self.patience = patience
        self.gpuid = str(gpuid)
        os.environ["CUDA_VISIBLE_DEVICES"]= self.gpuid        
        
        self.verbose = verbose
        self.random_state = random_state
        self.is_fit = False        
        self.name = "AggMap MultiLabels Estimator"
        
        #print(self.get_params())
        self.history = {}
        self.history_model = {}        
        
        if self.verbose > 0:
            self.verbose1 = self.verbose
            self.verbose2 = 0
        elif self.verbose ==0:
            self.verbose1 = 0
            self.verbose2 = 0
        elif self.verbose < 0:
            self.verbose1 = 0
            self.verbose2 = abs(self.verbose)
        print(self)
        
    def get_params(self, deep=True):

        model_paras =  {"epochs": self.epochs, 
                        "lr":self.lr, 
                        "loss":self.loss,
                        "conv1_kernel_size": self.conv1_kernel_size,
                        "dense_layers": self.dense_layers, 
                        "dense_avf":self.dense_avf, 
                        "batch_size":self.batch_size, 
                        "dropout":self.dropout,
                        "batch_norm":self.batch_norm,
                        "n_inception":self.n_inception,                        
                        "monitor": self.monitor,
                        "metric":self.metric,
                        "patience":self.patience,
                        "random_state":self.random_state,
                        "verbose":self.verbose,
                        "name":self.name,
                        "gpuid": self.gpuid,
                       }

        return model_paras
    
    
    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self

        

    def fit(self, X, y,  X_valid = None, y_valid = None):

        # Check that X and y have correct shape
        
        if  X.ndim != 4:
            raise ValueError("Found array X with dim %d. %s expected == 4." % (X.ndim, self.name))
            
        if  y.ndim != 2:
            raise ValueError("Found array y with dim %d. %s expected == 2." % (y.ndim, self.name))    
    
        # Store the classes seen during fit
        self.classes_ = unique_labels(y)

        self.X_ = X
        self.y_ = y
        
        if (X_valid is None) | (y_valid is None):
            
            X_valid = X
            y_valid = y
        
        np.random.seed(self.random_state)
        tf.compat.v1.set_random_seed(self.random_state)

            
        model = aggmodel.net._AggMapNet(X.shape[1:],
                                        n_outputs = y.shape[-1], 
                                        conv1_kernel_size = self.conv1_kernel_size,
                                        batch_norm = self.batch_norm,
                                        n_inception = self.n_inception,
                                        dense_layers = self.dense_layers, 
                                        dense_avf = self.dense_avf, 
                                        dropout = self.dropout,
                                        last_avf = None)
        
        opt = tf.keras.optimizers.Adam(lr=self.lr, beta_1=0.9, beta_2=0.999, epsilon=1e-08, decay=0.0) #
        model.compile(optimizer = opt, loss = self.loss)
        performance = aggmodel.cbks.CLA_EarlyStoppingAndPerformance((X, y), 
                                                                    (X_valid, y_valid), 
                                                                    patience = self.patience, 
                                                                    criteria = self.monitor,
                                                                    metric = self.metric,  
                                                                    last_avf = None,
                                                                    verbose = self.verbose1,)

        history = model.fit(X, y, 
                  batch_size=self.batch_size, 
                  epochs= self.epochs, verbose= self.verbose2, shuffle = True, 
                  validation_data = (X_valid, y_valid), 
                  callbacks=[performance]) 

        self._model = model
        self._performance = performance
        self.history = self._performance.history
        self.history_model = history.history
        self.is_fit = True
        
        return self



    def predict_proba(self, X):
        """
        Probability estimates.
        The returned estimates for all classes are ordered by the
        label of classes.
        For a multi_class problem, if multi_class is set to be "multinomial"
        the softmax function is used to find the predicted probability of
        each class.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Vector to be scored, where `n_samples` is the number of samples and
            `n_features` is the number of features.
        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Returns the probability of the sample for each class in the model,
            where classes are ordered as they are in ``self.classes_``.
        """
        # Check is fit had been called
        check_is_fitted(self)
        
        # Input validation
        if  X.ndim != 4:
            raise ValueError("Found array X with dim %d. %s expected == 4." % (X.ndim, self.name))
        y_prob = self._performance.sigmoid(self._model.predict(X))
        return y_prob
    

    def predict(self, X):
        y_pred = np.round(self.predict_proba(X))
        return y_pred
    
    

    def score(self, X, y, scoring = 'accuracy', sample_weight=None):
        """Returns the score using the `scoring` option on the given
        test data and labels.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Test samples.
        y : array-like of shape (n_samples,)
            True labels for X.
        scoring: str, please refer to: https://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter
        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights.
        Returns
        -------
        score : float
            Score of self.predict(X) wrt. y.
        """
        assert scoring in SCORERS.keys(), 'scoring is not in %s' % SCORERS.keys()
        scoring = get_scorer(scoring)

        return scoring(self, X, y, sample_weight=sample_weight)
    
    def plot_model(self, to_file='model.png', 
                   show_shapes=True, 
                   show_layer_names=True, 
                   rankdir='TB', 
                   expand_nested=False, 
                   dpi=96):
        if self.is_fit:
            tf.keras.utils.plot_model(self._model, 
                       to_file=to_file, 
                       show_shapes=show_shapes, 
                       show_layer_names=show_layer_names, 
                       rankdir=rankdir, 
                       expand_nested=expand_nested, 
                       dpi=dpi)
        else:
            print('Please fit first!')
            

    def save_model(self, model_path):
        return save_model(self, model_path)

    
    def load_model(self, model_path, gpuid=None):
        return load_model(model_path, gpuid=gpuid)
    
    
    def explain_model(self, mp, 
                      X, 
                      y, 
                      explain_format = 'global',
                      apply_logrithm = False, 
                      apply_smoothing = False, 
                      kernel_size = 3, sigma = 1.2):
        '''
        Feature importance calculation
        Parameters
        --------------
        mp: aggmap object
        X: trianing or test set X arrays
        y: trianing or test set y arrays
            whether the task is binary, if True, the feature importance will be calculated for one class only
        explain_format: {'local', 'global'}, default: 'global'
            local or global feature importance, if local, then X must be one sample
        apply_logrithm: {True, False}, default: False
            whether apply a logarithm transformation on the importance values
        apply_smoothing: {True, False}, default: False
            whether apply a smoothing transformation on the importance values
        kernel_size: odd number, the kernel size to perform the smoothing
        sigma: float, sigma for gaussian smoothing
        
        Returns
        ------------        
        DataFrame of feature importance
        '''
        if explain_format == 'global':
            explain_func = GlobalIMP
        else:
            explain_func = LocalIMP
        
        dfe = explain_func(self, mp, X, y, 
                           task_type = 'classification',
                           binary_task = False,
                           sigmoidy = True,  
                           apply_logrithm = apply_logrithm, 
                           apply_smoothing = apply_smoothing,
                           kernel_size = kernel_size, sigma = sigma)
        return dfe    
    
    @property
    def clean(self):
        clean(self)    
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Aug 25 20:29:36 2019

@author: wanxiang.shen@u.nus.edu

main aggmap code


"""
from aggmap.utils.logtools import print_info, print_warn, print_error
from aggmap.utils.matrixopt import Scatter2Grid, Scatter2Array, smartpadding 
from aggmap.utils import vismap, summary, calculator

from sklearn.cluster import AgglomerativeClustering
from sklearn.manifold import TSNE, MDS
from sklearn.utils import shuffle
from joblib import Parallel, delayed, load, dump
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import fcluster, linkage, dendrogram
import matplotlib.pylab as plt
import seaborn as sns
from umap import UMAP
from tqdm import tqdm
from copy import copy
import pandas as pd
import numpy as np
import os


class Base:
    
    def __init__(self):
        pass
        
    def _save(self, filename):
        return dump(self, filename)
        
    def _load(self, filename):
        return load(filename)

    
    def MinMaxScaleClip(self, x, xmin, xmax):
        scaled = (x - xmin) / ((xmax - xmin) + 1e-8)
        return scaled

    def StandardScaler(self, x, xmean, xstd):
        return (x-xmean) / (xstd + 1e-8) 
    

    
def _cluster_model2linkage_matrix(model):    
    counts = np.zeros(model.children_.shape[0])
    n_samples = len(model.labels_)
    for i, merge in enumerate(model.children_):
        current_count = 0
        for child_idx in merge:
            if child_idx < n_samples:
                current_count += 1  # leaf node
            else:
                current_count += counts[child_idx - n_samples]
        counts[i] = current_count

    linkage_matrix = np.column_stack([model.children_, model.distances_,
                                      counts]).astype(float)
    return linkage_matrix


class AggMap(Base):
    
    
    '''
    Note: t-SNE initialize method should be changed into 'pca': https://www.nature.com/articles/s41587-020-00809-z
    
    >>> mp = AggMap(dfx)
    >>> mp.fit(emb_method = 'tsne', init = 'pca')
    
    '''
    
    def __init__(self, 
                 dfx,
                 metric = 'correlation',
                ):
        
        """
        paramters
        -----------------
        dfx: pandas DataFrame
        metric: {'cosine', 'correlation', 'euclidean', 'jaccard', 'hamming', 'dice'}, default: 'correlation', measurement of feature distance

        """
        
        assert type(dfx) == pd.core.frame.DataFrame, 'input dfx mush be pandas DataFrame!'
        super().__init__()

        self.metric = metric
        self.isfit = False
        self.alist = dfx.columns.tolist()
        self.ftype = 'feature points'

        
        ## calculating distance
        print_info('Calculating distance ...')
        D = calculator.pairwise_distance(dfx.values, n_cpus=16, method=metric)
        D = np.nan_to_num(D,copy=False)
        D_ = squareform(D)
        self.info_distance = D_.clip(0, np.inf)

        ## statistic info
        S = summary.Summary(n_jobs = 10)
        res= []
        for i in tqdm(range(dfx.shape[1]), ascii=True):
            r = S._statistics_one(dfx.values, i)
            res.append(r)
        dfs = pd.DataFrame(res, index = self.alist)
        self.info_scale = dfs
        
        print_info('Applying the Agglomerative Clustering ...')
        cluster_model = AgglomerativeClustering(distance_threshold=0, n_clusters=None)
        cluster_model.fit(dfx.values.T)
        self._intrinsic_Z = _cluster_model2linkage_matrix(cluster_model)        
        
        
        
    def _fit_embedding(self, 
                        dist_matrix,
                        method = 'umap',  
                        n_components = 2,
                        random_state = 32,  
                        verbose = 2,
                        n_neighbors = 15,
                        min_dist = 0.1,
                        **kwargs):
        
        """
        parameters
        -----------------
        dist_matrix: distance matrix to fit
        method: {'tsne', 'umap', 'mds'}, algorithm to embedd high-D to 2D
        kwargs: the extra parameters for the conresponding algorithm
        """

        if 'metric' in kwargs.keys():
            metric = kwargs.get('metric')
            kwargs.pop('metric')
            
        else:
            metric = 'precomputed'

        if method == 'tsne':
            embedded = TSNE(n_components=n_components, 
                            random_state=random_state,
                            metric = metric,
                            verbose = verbose,
                            **kwargs)
        elif method == 'umap':
            embedded = UMAP(n_components = n_components, 
                            n_neighbors = n_neighbors,
                            min_dist = min_dist,
                            verbose = verbose,
                            random_state=random_state, 
                            metric = metric, **kwargs)
            
        elif method =='mds':
            if 'metric' in kwargs.keys():
                kwargs.pop('metric')
            if 'dissimilarity' in kwargs.keys():
                dissimilarity = kwargs.get('dissimilarity')
                kwargs.pop('dissimilarity')
            else:
                dissimilarity = 'precomputed'
                
            embedded = MDS(metric = True, 
                           n_components= n_components,
                           verbose = verbose,
                           dissimilarity = dissimilarity, 
                           random_state = random_state, **kwargs)
        
        embedded = embedded.fit(dist_matrix)    
        
        return embedded
    
    
   
            

    def fit(self, 
            feature_group_list = [],
            cluster_channels = 5,
            var_thr = -1, 
            split_channels = True, 
            fmap_type = 'grid',  
            fmap_shape = None, 
            emb_method = 'umap', 
            min_dist = 0.1, 
            n_neighbors = 15,
            verbose = 2, 
            random_state = 32,
            group_color_dict  = {},
            lnk_method = 'ward',
            **kwargs): 
        """
        parameters
        -----------------
        feature_group_list: list of the group name for each feature point
        cluster_channels: int, number of the channels(clusters) if feature_group_list is empty
        var_thr: float, defalt is -1, meaning that feature will be included only if the conresponding variance larger than this value. Since some of the feature has pretty low variances, we can remove them by increasing this threshold
        split_channels: bool, if True, outputs will split into various channels using the types of feature
        fmap_type:{'scatter', 'grid'}, default: 'gird', if 'scatter', will return a scatter mol map without an assignment to a grid
        fmap_shape: None or tuple, size of molmap, only works when fmap_type is 'scatter', if None, the size of feature map will be calculated automatically
        emb_method: {'tsne', 'umap', 'mds'}, algorithm to embedd high-D to 2D
        group_color_dict: dict of the group colors, keys are the group names, values are the colors
        lnk_method: {'ward','complete', 'average', 'single'}, linkage method
        kwargs: the extra parameters for the conresponding embedding method
        """
            
        if 'n_components' in kwargs.keys():
            kwargs.pop('n_components')
            
            
        ## embedding  into a 2d 
        assert emb_method in ['tsne', 'umap', 'mds'], 'No Such Method Supported: %s' % emb_method
        assert fmap_type in ['scatter', 'grid'], 'No Such Feature Map Type Supported: %s'   % fmap_type     
        self.var_thr = var_thr
        self.split_channels = split_channels
        self.fmap_type = fmap_type
        self.fmap_shape = fmap_shape
        self.emb_method = emb_method
        self.lnk_method = lnk_method
        
        if fmap_shape != None:
            assert len(fmap_shape) == 2, "fmap_shape must be a tuple with two elements!"
        
        # flist and distance
        flist = self.info_scale[self.info_scale['var'] > self.var_thr].index.tolist()
        
        dfd = pd.DataFrame(squareform(self.info_distance),
                           index=self.alist,
                           columns=self.alist)
        dist_matrix = dfd.loc[flist][flist]
        self.flist = flist
        
        self.x_mean = self.info_scale['mean'].values
        self.x_std =  self.info_scale['std'].values
        
        self.x_min = self.info_scale['min'].values
        self.x_max = self.info_scale['max'].values
        
   
        #bitsinfo
        dfb = pd.DataFrame(self.alist, columns = ['IDs'])
        if feature_group_list != []:

            self.Z = self._intrinsic_Z
            
            assert len(feature_group_list) == len(self.alist), "the length of the input group list is not equal to length of the feature list"
            self.cluster_channels = len(set(feature_group_list))
            self.feature_group_list = feature_group_list
            
            dfb['Subtypes'] = feature_group_list
            
            if set(feature_group_list).issubset(set(group_color_dict.keys())):
                self.group_color_dict = group_color_dict
                dfb['colors'] = dfb['Subtypes'].map(group_color_dict)
            else:
                unique_types = dfb['Subtypes'].unique()
                color_list = sns.color_palette("hsv", len(unique_types)).as_hex()
                group_color_dict = dict(zip(unique_types, color_list))
                dfb['colors'] = dfb['Subtypes'].map(group_color_dict)
                self.group_color_dict = group_color_dict
        else:
            
            self.cluster_channels = cluster_channels
            print_info('applying hierarchical clustering to obtain group information ...')

            if self.lnk_method != 'ward':
                Z = linkage(squareform(dfd.values),  lnk_method)
            else:
                Z = self._intrinsic_Z
                
            labels = fcluster(Z, cluster_channels, criterion='maxclust')
            
            feature_group_list = ['cluster_%s' % str(i).zfill(2) for i in labels]
            dfb['Subtypes'] = feature_group_list
            dfb = dfb.sort_values('Subtypes')
            unique_types = dfb['Subtypes'].unique()
            
            if not set(unique_types).issubset(set(group_color_dict.keys())):
                color_list = sns.color_palette("hsv", len(unique_types)).as_hex()
                group_color_dict = dict(zip(unique_types, color_list))
            
            dfb['colors'] = dfb['Subtypes'].map(group_color_dict)
            self.group_color_dict = group_color_dict           
            self.Z = Z
            self.feature_group_list = feature_group_list
            

        self.bitsinfo = dfb
        colormaps = dfb.set_index('Subtypes')['colors'].to_dict()
        colormaps.update({'NaN': '#000000'})
        self.colormaps = colormaps
  
        if fmap_type == 'grid':
            S = Scatter2Grid()
        else:
            if fmap_shape == None:
                N = len(self.flist)
                l = np.int(np.sqrt(N))*2
                fmap_shape = (l, l)                
            S = Scatter2Array(fmap_shape)
        
        self._S = S

        ## 2d embedding first
        embedded = self._fit_embedding(dist_matrix,
                                       method = emb_method,
                                       n_neighbors = n_neighbors,
                                       random_state = random_state,
                                       min_dist = min_dist, 
                                       verbose = verbose,
                                       n_components = 2, **kwargs)
        
        self.embedded = embedded 
        
        df = pd.DataFrame(embedded.embedding_, index = self.flist,columns=['x', 'y'])
        typemap = self.bitsinfo.set_index('IDs')
        df = df.join(typemap)
        df['Channels'] = df['Subtypes']
        self.df_embedding = df
      
        if self.fmap_type == 'scatter':
            ## naive scatter algorithm
            print_info('Applying naive scatter feature map...')
            self._S.fit(self.df_embedding, self.split_channels, channel_col = 'Channels')
            print_info('Finished')
            
        else:
            ## linear assignment algorithm 
            print_info('Applying grid feature map(assignment), this may take several minutes(1~30 min)')
            self._S.fit(self.df_embedding, self.split_channels, channel_col = 'Channels')
            print_info('Finished')
        
        ## fit flag
        self.isfit = True
        if self.fmap_shape == None:
            self.fmap_shape = self._S.fmap_shape        
        else:
            m, n = self.fmap_shape
            p, q = self._S.fmap_shape
            assert (m >= p) & (n >=q), "fmap_shape's width must >= %s, height >= %s " % (p, q)
        return self
        

    def transform(self, 
                  arr_1d, 
                  scale = True, 
                  scale_method = 'minmax',
                  fillnan = 0):
    
    
        """
        parameters
        --------------------
        arr_1d: 1d numpy array feature points
        scale: bool, if True, we will apply MinMax scaling by the precomputed values
        scale_method: {'minmax', 'standard'}
        fillnan: fill nan value, default: 0
        """
        
        if not self.isfit:
            print_error('please fit first!')
            return

        if scale:
            if scale_method == 'standard':
                arr_1d = self.StandardScaler(arr_1d, self.x_mean, self.x_std)
            else:
                arr_1d = self.MinMaxScaleClip(arr_1d, self.x_min, self.x_max)
        
        df = pd.DataFrame(arr_1d).T
        df.columns = self.alist

        df = df[self.flist]
        vector_1d = df.values[0] #shape = (N, )
        fmap = self._S.transform(vector_1d)  
        p, q, c = fmap.shape
        
        if self.fmap_shape != None:        
            m, n = self.fmap_shape
            if (m > p) | (n > q):
                fps = []
                for i in range(c):
                    fp = smartpadding(fmap[:,:,i], self.fmap_shape)
                    fps.append(fp)
                fmap = np.stack(fps, axis=-1)        

        return np.nan_to_num(fmap, nan = fillnan)   
    
    

    
    def batch_transform(self, 
                        array_2d, 
                        scale = True, 
                        scale_method = 'minmax',
                        n_jobs=4,
                        fillnan = 0):
    
        """
        parameters
        --------------------
        array_2d: 2D numpy array feature points, M(samples) x N(feature ponits)
        scale: bool, if True, we will apply MinMax scaling by the precomputed values
        scale_method: {'minmax', 'standard'}
        n_jobs: number of parallel
        fillnan: fill nan value, default: 0
        """
        
        if not self.isfit:
            print_error('please fit first!')
            return
        
        assert type(array_2d) == np.ndarray, 'input must be numpy ndarray!' 
        assert array_2d.ndim == 2, 'input must be 2-D  numpy array!' 
        
        P = Parallel(n_jobs=n_jobs)
        res = P(delayed(self.transform)(arr_1d, 
                                        scale,
                                        scale_method,
                                        fillnan) for arr_1d in tqdm(array_2d, ascii=True)) 
        X = np.stack(res) 
        
        return X
    
    
    def plot_scatter(self, htmlpath='./', htmlname=None, radius = 2, enabled_data_labels = False):
        """radius: the size of the scatter, must be int"""
        df_scatter, H_scatter = vismap.plot_scatter(self,  
                                                    htmlpath=htmlpath, 
                                                    htmlname=htmlname,
                                                    radius = radius,
                                                    enabled_data_labels = enabled_data_labels)
        
        self.df_scatter = df_scatter
        return H_scatter   
        
        
    def plot_grid(self, htmlpath='./', htmlname=None, enabled_data_labels = False):
        
        if self.fmap_type != 'grid':
            return
        
        df_grid, H_grid = vismap.plot_grid(self,
                                           htmlpath=htmlpath, 
                                           htmlname=htmlname,
                                           enabled_data_labels = enabled_data_labels)
        
        self.df_grid = df_grid
        return H_grid       
        
        
        
    def plot_tree(self, figsize=(16,8), add_leaf_labels = True, leaf_font_size = 18, leaf_rotation = 90):

        fig = plt.figure(figsize=figsize)
     
        Z = self.Z

        D_leaf_colors = self.bitsinfo['colors'].to_dict() 
        link_cols = {}
        for i, i12 in enumerate(Z[:,:2].astype(int)):
            c1, c2 = (link_cols[x] if x > len(Z) else D_leaf_colors[x] for x in i12)
            link_cols[i+1+len(Z)] = c1

        if add_leaf_labels:
            labels = self.alist
        else:
            labels = None

        P =dendrogram(Z, labels = labels, 
                      leaf_rotation = leaf_rotation, 
                      leaf_font_size = leaf_font_size, 
                      link_color_func=lambda x: link_cols[x])
        
        return fig
        
        
    def copy(self):
        return copy(self)
        
        
    def load(self, filename):
        return self._load(filename)
    
    
    def save(self, filename):
        return self._save(filename)


from sklearn.metrics import roc_auc_score, precision_recall_curve
from sklearn.metrics import auc as calculate_auc
from sklearn.metrics import mean_squared_error
from sklearn.metrics import accuracy_score
import tensorflow as tf
import os
import numpy as np


from scipy.stats.stats import pearsonr
def r2_score(x,y):
    pcc, _ = pearsonr(x,y)
    return pcc**2

def prc_auc_score(y_true, y_score):
    precision, recall, threshold  = precision_recall_curve(y_true, y_score) #PRC_AUC
    auc = calculate_auc(recall, precision)
    return auc

'''
for early-stopping techniques in regression and classification task
'''
######## Regression ###############################


class Reg_EarlyStoppingAndPerformance(tf.keras.callbacks.Callback):

    def __init__(self, train_data, valid_data, MASK = -1, patience=5, criteria = 'val_loss', verbose = 0):
        super(Reg_EarlyStoppingAndPerformance, self).__init__()
        
        assert criteria in ['val_loss', 'val_r2'], 'not support %s ! only %s' % (criteria, ['val_loss', 'val_r2'])
        self.x, self.y  = train_data
        self.x_val, self.y_val = valid_data
        
        self.history = {'loss':[],
                        'val_loss':[],
                        
                        'rmse':[],
                        'val_rmse':[],
                        
                        'r2':[],
                        'val_r2':[],
                        
                        'epoch':[]}
        self.MASK = MASK
        self.patience = patience
        # best_weights to store the weights at which the minimum loss occurs.
        self.best_weights = None
        self.criteria = criteria
        self.best_epoch = 0
        self.verbose = verbose
        
    def rmse(self, y_true, y_pred):

        N_classes = y_pred.shape[1]
        rmses = []
        for i in range(N_classes):
            y_pred_one_class = y_pred[:,i]
            y_true_one_class = y_true[:, i]
            mask = ~(y_true_one_class == self.MASK)
            mse = mean_squared_error(y_true_one_class[mask], y_pred_one_class[mask])
            rmse = np.sqrt(mse)
            rmses.append(rmse)
        return rmses   
    
    
    def r2(self, y_true, y_pred):
        N_classes = y_pred.shape[1]
        r2s = []
        for i in range(N_classes):
            y_pred_one_class = y_pred[:,i]
            y_true_one_class = y_true[:, i]
            mask = ~(y_true_one_class == self.MASK)
            r2 = r2_score(y_true_one_class[mask], y_pred_one_class[mask])
            r2s.append(r2)
        return r2s   
    
        
    def on_train_begin(self, logs=None):
        # The number of epoch it has waited when loss is no longer minimum.
        self.wait = 0
        # The epoch the training stops at.
        self.stopped_epoch = 0
        # Initialize the best as infinity.
        if self.criteria == 'val_loss':
            self.best = np.Inf  
        else:
            self.best = -np.Inf
            
        
        
 
        
    def on_epoch_end(self, epoch, logs={}):
        
        y_pred = self.model.predict(self.x)
        rmse_list = self.rmse(self.y, y_pred)
        rmse_mean = np.nanmean(rmse_list)
        
        r2_list = self.r2(self.y, y_pred) 
        r2_mean = np.nanmean(r2_list)
        
        
        y_pred_val = self.model.predict(self.x_val)
        rmse_list_val = self.rmse(self.y_val, y_pred_val)        
        rmse_mean_val = np.nanmean(rmse_list_val)
        
        r2_list_val = self.r2(self.y_val, y_pred_val)       
        r2_mean_val = np.nanmean(r2_list_val)        
        
        self.history['loss'].append(logs.get('loss'))
        self.history['val_loss'].append(logs.get('val_loss'))
        
        self.history['rmse'].append(rmse_mean)
        self.history['val_rmse'].append(rmse_mean_val)
        
        self.history['r2'].append(r2_mean)
        self.history['val_r2'].append(r2_mean_val)        
        
        self.history['epoch'].append(epoch)
        
        
        # logs is a dictionary
        eph = str(epoch+1).zfill(4)   
        loss = '{0:.4f}'.format((logs.get('loss')))
        val_loss = '{0:.4f}'.format((logs.get('val_loss')))
        rmse = '{0:.4f}'.format(rmse_mean)
        rmse_val = '{0:.4f}'.format(rmse_mean_val)
        r2_mean = '{0:.4f}'.format(r2_mean)
        r2_mean_val = '{0:.4f}'.format(r2_mean_val)
        
        if self.verbose:
            print('\repoch: %s, loss: %s - val_loss: %s; rmse: %s - rmse_val: %s;  r2: %s - r2_val: %s' % (eph,
                                                                                                           loss, val_loss, 
                                                                                                           rmse,rmse_val,
                                                                                                           r2_mean,r2_mean_val),
                  end=100*' '+'\n')


        if self.criteria == 'val_loss':
            current = logs.get(self.criteria)
            if current <= self.best:
                self.best = current
                self.wait = 0
                # Record the best weights if current results is better (less).
                self.best_weights = self.model.get_weights()
                self.best_epoch = epoch

            else:
                self.wait += 1
                if self.wait >= self.patience:
                    self.stopped_epoch = epoch
                    self.model.stop_training = True
                    print('\nRestoring model weights from the end of the best epoch.')
                    self.model.set_weights(self.best_weights)    
                    
        else:
            current = np.nanmean(r2_list_val)
            
            if current >= self.best:
                self.best = current
                self.wait = 0
                # Record the best weights if current results is better (less).
                self.best_weights = self.model.get_weights()
                self.best_epoch = epoch

            else:
                self.wait += 1
                if self.wait >= self.patience:
                    self.stopped_epoch = epoch
                    self.model.stop_training = True
                    print('\nRestoring model weights from the end of the best epoch.')
                    self.model.set_weights(self.best_weights)              
    
    def on_train_end(self, logs=None):
        self.model.set_weights(self.best_weights)
        if self.stopped_epoch > 0:
            print('\nEpoch %05d: early stopping' % (self.stopped_epoch + 1))

        
        
    def evaluate(self, testX, testY):
        """evalulate, return rmse and r2"""
        y_pred = self.model.predict(testX)
        rmse_list = self.rmse(testY, y_pred)
        r2_list = self.r2(testY, y_pred)
        return rmse_list, r2_list       



    
    
######## classification ###############################

class CLA_EarlyStoppingAndPerformance(tf.keras.callbacks.Callback):

    def __init__(self, train_data, valid_data, MASK = -1, patience=5, criteria = 'val_loss', metric = 'ROC', last_avf = None, verbose = 0):
        super(CLA_EarlyStoppingAndPerformance, self).__init__()
        
        sp = ['val_loss', 'val_metric']
        assert criteria in sp, 'not support %s ! only %s' % (criteria, sp)
        ms = ['ROC', 'PRC', 'ACC']
        assert metric in ms, 'not support %s ! only %s' % (metric, ms)        
        ms_dict = {'ROC':'roc_auc', 'PRC':'prc_auc', 'ACC':'accuracy'}
        
        metric = ms_dict[metric]
        val_metric = 'val_%s' % metric
        self.metric = metric
        self.val_metric = val_metric
        
        self.x, self.y  = train_data
        self.x_val, self.y_val = valid_data
        self.last_avf = last_avf
        
        
        
        self.history = {'loss':[],
                        'val_loss':[],
                        self.metric:[],
                        self.val_metric:[],
                        'epoch':[]}
        self.MASK = MASK
        self.patience = patience
        # best_weights to store the weights at which the minimum loss occurs.
        self.best_weights = None
        self.criteria = criteria

        
        self.best_epoch = 0
        self.verbose = verbose
        
    def sigmoid(self, x):
        s = 1/(1+np.exp(-x))
        return s

    
    def roc_auc(self, y_true, y_pred):
        if self.last_avf == None:
            y_pred_logits = self.sigmoid(y_pred)
        else:
            y_pred_logits = y_pred
            
        N_classes = y_pred_logits.shape[1]

        aucs = []
        for i in range(N_classes):
            y_pred_one_class = y_pred_logits[:,i]
            y_true_one_class = y_true[:, i]
            mask = ~(y_true_one_class == self.MASK)
            try:
                if self.metric == 'roc_auc':
                    auc = roc_auc_score(y_true_one_class[mask], y_pred_one_class[mask], average='weighted') #ROC_AUC
                elif self.metric == 'prc_auc': 
                    auc = prc_auc_score(y_true_one_class[mask], y_pred_one_class[mask]) #PRC_AUC
                elif self.metric == 'accuracy':
                    auc = accuracy_score(y_true_one_class[mask], np.round(y_pred_one_class[mask])) #ACC
            except:
                auc = np.nan
            aucs.append(auc)
        return aucs  
    
        
        
    def on_train_begin(self, logs=None):
        # The number of epoch it has waited when loss is no longer minimum.
        self.wait = 0
        # The epoch the training stops at.
        self.stopped_epoch = 0
        # Initialize the best as infinity.
        if self.criteria == 'val_loss':
            self.best = np.Inf  
        else:
            self.best = -np.Inf

            
    def on_epoch_end(self, epoch, logs={}):
        
        y_pred = self.model.predict(self.x)
        roc_list = self.roc_auc(self.y, y_pred)
        roc_mean = np.nanmean(roc_list)
        
        y_pred_val = self.model.predict(self.x_val)
        roc_val_list = self.roc_auc(self.y_val, y_pred_val)        
        roc_val_mean = np.nanmean(roc_val_list)
        
        self.history['loss'].append(logs.get('loss'))
        self.history['val_loss'].append(logs.get('val_loss'))
        self.history[self.metric].append(roc_mean)
        self.history[self.val_metric].append(roc_val_mean)
        self.history['epoch'].append(epoch)
        
        eph = str(epoch+1).zfill(4)        
        loss = '{0:.4f}'.format((logs.get('loss')))
        val_loss = '{0:.4f}'.format((logs.get('val_loss')))
        auc = '{0:.4f}'.format(roc_mean)
        auc_val = '{0:.4f}'.format(roc_val_mean)    
        
        if self.verbose:
            print('\repoch: %s, loss: %s - val_loss: %s; %s: %s - %s: %s' % (eph,
                                                                             loss, 
                                                                             val_loss, 
                                                                             self.metric,
                                                                             auc,
                                                                             self.val_metric,
                                                                             auc_val), end=100*' '+'\n')

        if self.criteria == 'val_loss':
            current = logs.get(self.criteria)
            if current <= self.best:
                self.best = current
                self.wait = 0
                # Record the best weights if current results is better (less).
                self.best_weights = self.model.get_weights()
                self.best_epoch = epoch

            else:
                self.wait += 1
                if self.wait >= self.patience:
                    self.stopped_epoch = epoch
                    self.model.stop_training = True
                    print('\nRestoring model weights from the end of the best epoch.')
                    self.model.set_weights(self.best_weights)    
                    
        else:
            current = roc_val_mean
            if current >= self.best:
                self.best = current
                self.wait = 0
                # Record the best weights if current results is better (less).
                self.best_weights = self.model.get_weights()
                self.best_epoch = epoch

            else:
                self.wait += 1
                if self.wait >= self.patience:
                    self.stopped_epoch = epoch
                    self.model.stop_training = True
                    print('\nRestoring model weights from the end of the best epoch.')
                    self.model.set_weights(self.best_weights)              
    
    def on_train_end(self, logs=None):
        self.model.set_weights(self.best_weights)
        if self.stopped_epoch > 0:
            print('\nEpoch %05d: early stopping' % (self.stopped_epoch + 1))

        
    def evaluate(self, testX, testY):
        
        y_pred = self.model.predict(testX)
        roc_list = self.roc_auc(testY, y_pred)
        return roc_list            

            



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 18 13:01:00 2020

@author: SHEN WANXIANG
"""



import warnings
warnings.filterwarnings("ignore")
import matplotlib.pyplot as plt
import seaborn as sns





def imshow(x_arr,  ax, mode = 'dark',  color_list = ['#1300ff','#ff0c00','#25ff00', '#d000ff','#e2ff00', 
                                                     '#00fff6', '#ff8800', '#fccde5','#178b66', '#8a0075'],
           x_max = 255, vmin = -1, vmax = 1,):
    
    
    assert x_arr.ndim == 3, 'input must be 3d array!'
    w, h, c = x_arr.shape
    assert len(color_list) >= c, 'length of the color list should equal or larger than channel numbers'
    
    x = x_arr.copy()
    x[x == 0] = 'nan'

    xxx = x_arr.sum(axis=-1)
    xxx[xxx != 0] = 'nan'

    if mode == 'dark':
        cmaps = [sns.dark_palette(color, n_colors =  50, reverse=False) for color in color_list]

    else:
        cmaps = [sns.light_palette(color, n_colors =  50, reverse=False) for color in color_list]
        
    for i in range(c):
        data = x[:,:,i]/x_max
        sns.heatmap(data, cmap = cmaps[i],  vmin = vmin, vmax = vmax,  
                    yticklabels=False, xticklabels=False, cbar=False, ax=ax, ) # linewidths=0.005, linecolor = '0.9'

    if mode == 'dark':
        sns.heatmap(xxx, vmin=-10000, vmax=1, cmap = 'Greys', yticklabels=False, xticklabels=False, cbar=False, ax=ax)
    else:
        sns.heatmap(xxx, vmin=0, vmax=1, cmap = 'Greys', yticklabels=False, xticklabels=False, cbar=False, ax=ax)
        ax.axhline(y=0, color='grey',lw=2, ls =  '--')
        ax.axhline(y=data.shape[0], color='grey',lw=2, ls =  '--')
        ax.autoscale()
        ax.axvline(x=data.shape[1], color='grey',lw=2, ls =  '--')  
        ax.axvline(x=0, color='grey',lw=2, ls =  '--')


def imshow_wrap(x,  mode = 'dark', color_list = ['#1300ff','#ff0c00','#25ff00', '#d000ff','#e2ff00', 
                                                 '#00fff6', '#ff8800', '#fccde5','#178b66', '#8a0075'], 
                x_max = 255, vmin = -1, vmax = 1,):
    
    fig, ax = plt.subplots(figsize=(4,4))
    imshow(x.astype(float),mode = mode, color_list = color_list, ax=ax, x_max = x_max, vmin = vmin, vmax=vmax)


import tensorflow as tf
from tensorflow.keras.utils import plot_model
from tensorflow.keras import Model, Input, layers
from tensorflow.keras.layers import MaxPool2D, GlobalMaxPool2D, BatchNormalization, Activation
from tensorflow.keras.layers import Conv2D, Concatenate,Flatten, Dense, Dropout
from tensorflow.keras.regularizers import l2, l1

#from tensorflow_addons.layers import  GroupNormalization

def count_trainable_params(model):
    p = 0
    for layer in model.layers:
        if len(layer.trainable_variables) == 0:
            continue
        else:
            for variables in layer.trainable_variables:
                a = variables.shape.as_list()
                if len(a) == 1:
                    p += a[0]
                else:
                    p += a[0]*a[1]
    return p


def resnet_block(input_data, filters, conv_size):
    x = layers.Conv2D(filters, conv_size, activation='relu', padding='same')(input_data)
    x = layers.BatchNormalization()(x)
    x = layers.Conv2D(filters, conv_size, activation=None, padding='same')(x)
    x = layers.BatchNormalization()(x)
    x = layers.Add()([x, input_data])
    x = layers.Activation('relu')(x)
    return x


def Inception(inputs, units = 8, strides = 1):
    """
    naive google inception block
    """
    x1 = Conv2D(units, 5, padding='same', activation = 'relu', strides = strides)(inputs)
    x2 = Conv2D(units, 3, padding='same', activation = 'relu', strides = strides)(inputs)
    x3 = Conv2D(units, 1, padding='same', activation = 'relu', strides = strides)(inputs)
    outputs = Concatenate()([x1, x2, x3])    
    return outputs


def _AggMapNet(input_shape,  
               n_outputs = 1, 
               conv1_kernel_size = 11,
               batch_norm = False,
               n_inception = 2,
               dense_layers = [128, 32], 
               dense_avf = 'relu', 
               dropout = 0,
               last_avf = None):

    """
    parameters
    ----------------------
    molmap_shape: w, h, c
    n_outputs: output units
    n_inception: number of the inception layers
    dense_layers: list, how many dense layers and units
    dense_avf: activation function for dense layers
    last_avf: activation function for last layer
    dropout: dropout of the dense layers
    """
    tf.keras.backend.clear_session()
    assert len(input_shape) == 3
    inputs = Input(input_shape)
    
    if batch_norm:
        conv1 = Conv2D(48,  conv1_kernel_size, padding = 'same', strides = 1)(inputs) #activation='relu',  use_bias=False, kernel_regularizer=l1(1e-3)     
        conv1 = BatchNormalization(axis=-1, epsilon=1e-5, momentum=0.8)(conv1)
        incept = Activation("relu")(conv1)    
    else:
        incept = Conv2D(48,  conv1_kernel_size, padding = 'same', activation='relu', strides = 1)(inputs)
    
    for i in range(n_inception):
        incept = MaxPool2D(pool_size = 3, strides = 2, padding = 'same')(incept) #p1
        incept = Inception(incept, strides = 1, units = 32*(2**i))

    #flatten
    x = GlobalMaxPool2D()(incept)
    
    ## dense layer
    for units in dense_layers:
        x = Dense(units, activation = dense_avf)(x)
        if dropout:
            x = Dropout(rate = dropout)(x)

    #last layer
    outputs = Dense(n_outputs,activation=last_avf)(x)
    
    model = tf.keras.Model(inputs=inputs, outputs=outputs)
    
    return model

def _AggMapDualPathNet(molmap1_size, 
                    molmap2_size, 
                    n_outputs = 1,
                    conv1_kernel_size = 11,
                    dense_layers = [256, 128, 32], 
                    dense_avf = 'relu', 
                    last_avf = None):
    """
    parameters
    ----------------------
    molmap1_size: w, h, c, shape of first molmap
    molmap2_size: w, h, c, shape of second molmap
    n_outputs: output units
    dense_layers: list, how many dense layers and units
    dense_avf: activation function for dense layers
    last_avf: activation function for last layer
    """
    tf.keras.backend.clear_session()
    ## first inputs
    d_inputs1 = Input(molmap1_size)
    d_conv1 = Conv2D(48, conv1_kernel_size, padding = 'same', activation='relu', strides = 1)(d_inputs1)
    d_pool1 = MaxPool2D(pool_size = 3, strides = 2, padding = 'same')(d_conv1) #p1
    d_incept1 = Inception(d_pool1, strides = 1, units = 32)
    d_pool2 = MaxPool2D(pool_size = 3, strides = 2, padding = 'same')(d_incept1) #p2
    d_incept2 = Inception(d_pool2, strides = 1, units = 64)
    d_flat1 = GlobalMaxPool2D()(d_incept2)

    
    ## second inputs
    f_inputs1 = Input(molmap2_size)
    f_conv1 = Conv2D(48, conv1_kernel_size, padding = 'same', activation='relu', strides = 1)(f_inputs1)
    f_pool1 = MaxPool2D(pool_size = 3, strides = 2, padding = 'same')(f_conv1) #p1
    f_incept1 = Inception(f_pool1, strides = 1, units = 32)
    f_pool2 = MaxPool2D(pool_size = 3, strides = 2, padding = 'same')(f_incept1) #p2
    f_incept2 = Inception(f_pool2, strides = 1, units = 64)
    f_flat1 = GlobalMaxPool2D()(f_incept2)    
    
    ## concat
    x = Concatenate()([d_flat1, f_flat1]) 
    
    ## dense layer
    for units in dense_layers:
        x = Dense(units, activation = dense_avf)(x)
        #x = BatchNormalization()(x)
        
    ## last layer
    outputs = Dense(n_outputs, activation=last_avf)(x)
    model = tf.keras.Model(inputs=[d_inputs1, f_inputs1], outputs=outputs)
    
    return model



def _AggMapAddPathNet(molmap_shape,  additional_shape,
                    n_outputs = 1,              
                    dense_layers = [128, 32], 
                    dense_avf = 'relu', 
                    last_avf = None):
    
    
    """
    parameters
    ----------------------
    molmap_shape: w, h, c
    n_outputs: output units
    dense_layers: list, how many dense layers and units
    dense_avf: activation function for dense layers
    last_avf: activation function for last layer
    """
    tf.keras.backend.clear_session()
    assert len(molmap_shape) == 3
    inputs = Input(molmap_shape)
    inputs_actvie_amount = Input(additional_shape)
    
    conv1 = Conv2D(48, 13, padding = 'same', activation='relu', strides = 1)(inputs)
    
    conv1 = MaxPool2D(pool_size = 3, strides = 2, padding = 'same')(conv1) #p1
    
    incept1 = Inception(conv1, strides = 1, units = 32)
    
    incept1 = MaxPool2D(pool_size = 3, strides = 2, padding = 'same')(incept1) #p2
    
    incept2 = Inception(incept1, strides = 1, units = 64)
    
    #flatten
    x = GlobalMaxPool2D()(incept2)
    x = tf.keras.layers.concatenate([x, inputs_actvie_amount], axis=-1)
    
    ## dense layer
    for units in dense_layers:
        x = Dense(units, activation = dense_avf)(x)
        #x = BatchNormalization()(x)
    #last layer
    outputs = Dense(n_outputs,activation=last_avf)(x)
    
    
    model = tf.keras.Model(inputs=[inputs, inputs_actvie_amount], outputs=outputs)
    
    return model



def _AggMapResNet(input_shape,
                 num_resnet_blocks = 8,
                n_outputs = 1, 
                dense_layers = [128, 32], 
                dense_avf = 'relu', 
                last_avf = None                
                ):
    
    
    """
    parameters
    ----------------------
    molmap_shape: w, h, c
    num_resnet_blocks: int
    n_outputs: output units
    dense_layers: list, how many dense layers and units
    dense_avf: activation function for dense layers
    last_avf: activation function for last layer
    """
    tf.keras.backend.clear_session()
    inputs = tf.keras.Input(input_shape) #input_shape = (24, 24, 3)
#     x = layers.Conv2D(32, 11, activation='relu')(inputs)
#     x = layers.Conv2D(64, 3, activation='relu')(x)
#     x = layers.MaxPooling2D(3)(x)
    x = Conv2D(64,  13, padding = 'same', activation='relu', strides = 1)(inputs)
    x = MaxPool2D(pool_size = 3, strides = 2, padding = 'same')(x) #p1
    
    ## renet block 
    for i in range(num_resnet_blocks):
        x = resnet_block(x, 64, 3)

    x = layers.Conv2D(256, 3, activation='relu')(x)
    x = layers.GlobalMaxPool2D()(x)

     ## dense layer
    for units in dense_layers:
        x = layers.Dense(units, activation = dense_avf)(x)
        #x = BatchNormalization()(x)
    #last layer
    outputs = Dense(n_outputs,activation=last_avf)(x)
    
    model = tf.keras.Model(inputs=inputs, outputs=outputs)

    return model







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 17 16:54:12 2019

@author: wanxiang.shen@u.nus.edu

@logtools
"""


import os, sys, logging, time, traceback, inspect
from colorlog import ColoredFormatter
from colored import fg, bg, attr

formatter = ColoredFormatter(
    "%(asctime)s - %(log_color)s%(levelname).4s%(reset)s - %(message_log_color)s[%(name)s]%(reset)s - %(message_log_color)s%(message)s",
    datefmt=None,
    reset=True,
    log_colors={
        'DEBUG': 'cyan',
        'INFO': 'green',
        'WARNING': 'yellow',
        'ERROR': 'red',
        'CRITICAL': 'red,bg_white',
    },
    secondary_log_colors={
        'message':{
            'ERROR': 'red',
            'CRITICAL': 'red,bg_white',
        },
    },
    style='%',
    )


logger = logging.getLogger('bidd-aggmap')
logger.propagate = False


all_loggers = [ logger]

def set_level(level):
    for _logger in all_loggers:
        _logger.setLevel(getattr(logging, level.upper()))

set_level('INFO')

file_handler = None
def log_to_file(path):
    global file_handler
    if file_handler is not None:
        for _logger in all_loggers:
            _logger.removeHandler(file_handler)
            
    logpath = os.path.join(os.getcwd(), path + '.' + get_datetime() + '.log')
    print_info('log to file:', logpath)
    file_handler = logging.FileHandler(logpath)
    file_handler.setFormatter(formatter)
    for _logger in all_loggers:
        _logger.addHandler(file_handler)

def reset_handler(handler):
    for _logger in all_loggers:
        del _logger.handlers[:]
        _logger.addHandler(handler)
        if file_handler is not None:
            _logger.addHandler(file_handler)

handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(formatter)
reset_handler(handler)

DEFAULT_TEXT_LENGTH = 5000
DEFAULT_TEXT_LENGTH_PREFIX = 4000
DEFAULT_TEXT_LENGTH_SUFFIX = DEFAULT_TEXT_LENGTH - DEFAULT_TEXT_LENGTH_PREFIX
def set_text_length(prefix, suffix):
    global DEFAULT_TEXT_LENGTH, DEFAULT_TEXT_LENGTH_PREFIX, DEFAULT_TEXT_LENGTH_SUFFIX
    DEFAULT_TEXT_LENGTH = prefix + suffix
    DEFAULT_TEXT_LENGTH_PREFIX = prefix
    DEFAULT_TEXT_LENGTH_SUFFIX = suffix

def clip_text(text):
    if len(text) > DEFAULT_TEXT_LENGTH:
        text = '%s %s%s... [%d chars truncated] ...%s %s' % (text[:DEFAULT_TEXT_LENGTH_PREFIX], fg('red'), attr('bold'), len(text) - DEFAULT_TEXT_LENGTH, attr('reset'), text[-DEFAULT_TEXT_LENGTH_SUFFIX:])
    return text

def create_print_method(level):
    print_method = getattr(logger, level)
    def func(*args, sep=' ', verbose=True):
        if verbose: print_method(clip_text(sep.join(map(str, args))))
    return func

print_error = create_print_method('error')
print_warn = create_print_method('warn')
print_info = create_print_method('info')
print_debug = create_print_method('debug')

def format_exc(error):
    return traceback.format_exception(error.__class__, error, error.__traceback__)

def print_exc(error, verbose=True):
    lines = format_exc(error)
    logger.error(lines[-1].rstrip())
    if verbose:
        logger.info(''.join(lines[:-1]))

def print_exc_s(error):
    logger.error('%s: %s', type(error).__name__, error)

def print_traceback():
    logger.info('Traceback (most recent call last):')
    for frame in reversed(inspect.getouterframes(inspect.currentframe())[1:]):
        logger.info('  File "%s", line %s, in %s', frame.filename, frame.lineno, frame.function)
        logger.info('    %s', frame.code_context[0].strip())

last_time = time.time()
def print_timedelta(*args, sep=' '):
    global last_time
    this_time = time.time()
    logger.info('[%7.2f] %s', (this_time - last_time) * 1000, sep.join(map(str, args)))
    last_time = this_time

def get_date():
    return time.strftime('%Y%m%d')

def get_datetime():
    return time.strftime('%Y%m%d%H%M%S')

from tqdm import tqdm

class PBarHandler(logging.Handler):

    def __init__(self, pbar):
        logging.Handler.__init__(self)
        self.pbar = pbar

    def emit(self, record):
        self.pbar.write(self.format(record))

def pbar(*args, **kwargs):
    kwargs['ascii'] = kwargs.get('ascii', True)
    kwargs['smoothing'] = kwargs.get('smoothing', 0.7)

    pb = tqdm(*args, **kwargs)
    handler = PBarHandler(pb)
    handler.setFormatter(formatter)

    pb.handler = handler
    reset_handler(handler)

    return pb

if __name__ == '__main__':
    pb = pbar(range(100))
    for i in pb:
        time.sleep(0.05)
        print_info(str(i))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Aug 25 20:29:36 2019

@author: wanxiang.shen@u.nus.edu

matrix operation

"""

import numpy as np
from lapjv import lapjv
from scipy.signal import convolve2d
from scipy.spatial.distance import cdist


class Scatter2Grid:
    
    def __init__(self):  
        """assign x,y coords to gird numpy array"""
        self.fmap_shape = None
        self.indices = None
        self.indices_list = None
        
    def fit(self, df, split_channels = True, channel_col = 'Channels'):
        """
        parameters
        ------------------
        df: dataframe with x, y columns
        split_channels: bool, if True, will apply split by group
        channel_col: column in df.columns, split to groups by this col        
        
        """
        df['idx'] = range(len(df))
        
        embedding_2d = df[['x','y']].values
        N = len(df)

        size1 = int(np.ceil(np.sqrt(N)))
        size2 = int(np.ceil(N/size1))
        grid_size = (size1, size2)
        
        grid = np.dstack(np.meshgrid(np.linspace(0, 1, size2), 
                                     np.linspace(0, 1, size1))).reshape(-1, 2)
        grid_map = grid[:N]
        cost_matrix = cdist(grid_map, embedding_2d, "sqeuclidean").astype(np.float)
        cost_matrix = cost_matrix * (100000 / cost_matrix.max())
        row_asses, col_asses, _ = lapjv(cost_matrix)

        self.row_asses = row_asses
        self.col_asses = col_asses
        self.fmap_shape = grid_size
        self.indices = col_asses

        self.channel_col = channel_col
        self.split_channels = split_channels
        df['indices'] = self.indices
        self.df = df
        
        if self.split_channels:
            def _apply_split(x):
                return x[['idx', 'indices']].to_dict('list')
            sidx = df.groupby(channel_col).apply(_apply_split)      
            channels = sidx.index.tolist()
            indices_list = sidx.tolist()            
            self.channels = channels
            self.indices_list = indices_list

            
    def refit_c(self, df):
        """
        parameters
        ------------------
        df: dataframe with x, y columns
      
        
        """
        df['idx'] = range(len(df))
        df['indices'] = self.indices
        self.df = df
        
        if self.split_channels:
            def _apply_split(x):
                return x[['idx', 'indices']].to_dict('list')
            sidx = df.groupby(self.channel_col).apply(_apply_split)      
            channels = sidx.index.tolist()
            indices_list = sidx.tolist()            
            self.channels = channels
            self.indices_list = indices_list
            
            
    def transform(self, vector_1d):
        """vector_1d: extracted features
        """             
        ### linear assignment map ###
        M, N = self.fmap_shape

        if self.split_channels:
            arr_res = []
            for idict in self.indices_list:

                indices = idict['indices']
                idx = idict['idx']

                arr = np.zeros(self.fmap_shape)
                arr_1d = arr.reshape(M*N, )
                arr_1d[indices] = vector_1d[idx]
                arr = arr_1d.reshape(M, N)  
                arr_res.append(arr) 
            arr_res = np.stack(arr_res, axis=-1)
        else:
            arr_res = np.zeros(self.fmap_shape)
            arr_1d = arr_res.reshape(M*N, )
            arr_1d[self.indices] = vector_1d
            arr_res = arr_1d.reshape(M, N, 1)          
        return arr_res
    

    
class Scatter2Array:
    
    def __init__(self, fmap_shape = (128,128)):  
        """convert x,y coords to numpy array"""
        self.fmap_shape = fmap_shape
        self.indices = None
        self.indices_list = None
        
    def _fit(self, df):
        """df: dataframe with x, y columns"""
        M, N = self.fmap_shape
        self.X = np.linspace(df.x.min(), df.x.max(), M)
        self.Y = np.linspace(df.y.min(), df.y.max(), N)

    
    def _transform(self, dfnew):
        """dfnew: dataframe with x, y columns
           in case we need to split channels
        """             
        x = dfnew.x.values
        y = dfnew.y.values
        M, N = self.fmap_shape
        indices = []
        for i in range(len(dfnew)):
            #perform a l1 distance
            idx = np.argmin(abs(self.X-x[i]))
            idy = np.argmin(abs(self.Y-y[i]))     
            indice = N*idy + idx
            indices.append(indice)
        return indices
    
    
    def fit(self, df, split_channels = True, channel_col = 'Channels'):
        """
        parameters
        ---------------
        df: embedding_df, dataframe
        split_channels: bool, if True, will apply split by group
        channel_col: column in df.columns, split to groups by this col
        """
        df['idx'] = range(len(df))
        self.df = df
        self.channel_col = channel_col
        self.split_channels = split_channels
        _ = self._fit(df)
        
        if self.split_channels:
            g = df.groupby(channel_col)
            sidx = g.apply(self._transform)            
            self.channels = sidx.index.tolist()
            self.indices_list = sidx.tolist()
        else:    
            self.indices = self._transform(df)
            
            
    def transform(self, vector_1d):
        """vector_1d: feature values 1d array"""
        
        M, N = self.fmap_shape
        arr = np.zeros(self.fmap_shape)
        arr_1d = arr.reshape(M*N, )
            
        if self.split_channels:
            df = self.df
            arr_res = []
            for indices, channel in zip(self.indices_list, self.channels):
                arr = np.zeros(self.fmap_shape)
                df1 = df[df[self.channel_col] == channel]
                idx = df1.idx.tolist()
                arr_1d_copy = arr_1d.copy()
                arr_1d_copy[indices] = vector_1d[idx]
                arr_1d_copy = arr_1d_copy.reshape(M, N) 
                arr_res.append(arr_1d_copy)
            arr_res = np.stack(arr_res, axis=-1)
        else:
            arr_1d_copy = arr_1d.copy()
            arr_1d_copy[self.indices] = vector_1d
            arr_res = arr_1d_copy.reshape(M, N, 1) 
        return arr_res


def smartpadding(array, target_size, mode='constant', constant_values=0):
    """
    array: 2d array to be padded
    target_size: tuple of target array's shape
    """
    X, Y = array.shape
    M, N = target_size
    top = int(np.ceil((M-X)/2))
    bottom = int(M - X - top)
    right = int(np.ceil((N-Y)/2))
    left = int(N - Y - right)
    array_pad = np.pad(array, pad_width=[(top, bottom),
                                         (left, right)], 
                       mode=mode, 
                       constant_values=constant_values)
    
    return array_pad


def fspecial_gauss(size = 31, sigma = 2):

    """Function to mimic the 'fspecial' gaussian MATLAB function
      size should be odd value
    """
    x, y = np.mgrid[-size//2 + 1:size//2 + 1, -size//2 + 1:size//2 + 1]
    g = np.exp(-((x**2 + y**2)/(2.0*sigma**2)))
    return g/g.sum()


def conv2(array, kernel_size = 31, sigma = 2,  mode='same', fillvalue = 0):
    kernel = fspecial_gauss(kernel_size, sigma)
    return np.rot90(convolve2d(np.rot90(array, 2), np.rot90(kernel, 2), 
                               mode=mode, 
                               fillvalue = fillvalue), 2)



# grad-cam example codes: https://keras.io/examples/vision/grad_cam/#the-gradcam-algorithm



def make_gradcam_heatmap(img_array, model, last_conv_layer_name, pred_index=None):
    # First, we create a model that maps the input image to the activations
    # of the last conv layer as well as the output predictions
    grad_model = tf.keras.models.Model(
        [model.inputs], [model.get_layer(last_conv_layer_name).output, model.output]
    )

    # Then, we compute the gradient of the top predicted class for our input image
    # with respect to the activations of the last conv layer
    with tf.GradientTape() as tape:
        last_conv_layer_output, preds = grad_model(img_array)
        if pred_index is None:
            pred_index = tf.argmax(preds[0])
        class_channel = preds[:, pred_index]

    # This is the gradient of the output neuron (top predicted or chosen)
    # with regard to the output feature map of the last conv layer
    grads = tape.gradient(class_channel, last_conv_layer_output)

    # This is a vector where each entry is the mean intensity of the gradient
    # over a specific feature map channel
    pooled_grads = tf.reduce_mean(grads, axis=(0, 1, 2))

    # We multiply each channel in the feature map array
    # by "how important this channel is" with regard to the top predicted class
    # then sum all the channels to obtain the heatmap class activation
    last_conv_layer_output = last_conv_layer_output[0]
    heatmap = last_conv_layer_output @ pooled_grads[..., tf.newaxis]
    heatmap = tf.squeeze(heatmap)

    # For visualization purpose, we will also normalize the heatmap between 0 & 1
    heatmap = tf.maximum(heatmap, 0) / tf.math.reduce_max(heatmap)
    return heatmap.numpy()


import tensorflow as tf
from sklearn.metrics import roc_auc_score
import numpy as np


###########  classification ##############
def cross_entropy(y_true, y_pred):
    cost = tf.nn.sigmoid_cross_entropy_with_logits(labels=labels, logits=logits)
    return cost


def weighted_cross_entropy(y_true, y_pred, pos_weight):
  
    cost = tf.nn.weighted_cross_entropy_with_logits(labels=labels, logits=logits, 
                                                    pos_weight = pos_weight)
    return cost





def MALE(y_obs, y_pred):
    return tf.keras.backend.log(0.5 + tf.keras.backend.abs(y_pred - y_obs))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 17 16:54:12 2019

@author: wanxiang.shen@u.nus.edu

@usecase: calculate varies distances
"""
from aggmap.utils.distances import named_distances
from aggmap.utils.multiproc import MultiProcessUnorderedBarRun

import numpy as np
from tqdm import tqdm


def _yield_combinations(N):
    for i1 in range(N):
        for i2 in range(i1):
            yield (i1,i2)
            
def _calculate(i1, i2):
    x1 = data[:, i1]
    x2 = data[:, i2]
    ## dropna
    X = np.stack([x1,x2], axis=1)
    X = X[~np.isnan(X).any(axis=1)]
    x1 = X[:, 0]
    x2 = X[:, 1]
    
#     x1 = np.nan_to_num(x1)
#     x2 = np.nan_to_num(x2)
    if x1.any():
        dist = func(x1, x2)
    else:
        dist = np.nan
    return (i1, i2, dist)

def _fuc(x):
    i1, i2  = x
    return _calculate(i1, i2)


def pairwise_distance(npydata, n_cpus=8, method='correlation'):
    """
    parameters
    ---------------
    method: {'euclidean', 'manhattan', 'canberra', 'chebyshev', 
             'cosine', 'braycurtis', 'correlation',
             'jaccard', 'rogerstanimoto', 'hamming', 'dice', 'kulsinski', 'sokal_sneath'}
    npydata: np.array or np.memmap, Note that the default we will calcuate the vector's distances instead of sample's distances, if you wish to calculate distances between samples, you can pass data.T instead of data

    Usage
    --------------
    >>> import numpy as np
    >>> data = np.random.random_sample(size=(10000,10)
    >>> dist_matrix = pairwise_distance(data)
    >>> dist_matrix.shape
    >>> (10,10)  
    """    
    global data, func
    
    func = named_distances.get(method)
    data = npydata
    N = data.shape[1]
    lst = list(_yield_combinations(N))
    res = MultiProcessUnorderedBarRun(_fuc, lst, n_cpus=n_cpus)
    dist_matrix = np.zeros(shape = (N,N))
    for x,y,v in tqdm(res,ascii=True):
        dist_matrix[x,y] = v
        dist_matrix[y,x] = v
    return dist_matrix

    
    
if __name__ == '__main__':
    
    import numpy as np
    import pandas as pd
    from umap import UMAP
    import matplotlib.pyplot as plt
    
    X = np.random.random_sample(size=(1000000,40))
    distmatrix = pairwise_distance(X, n_cpus=6)
    embedding = UMAP(metric = 'precomputed',random_state=10)
    df = pd.DataFrame(embedding.fit_transform(distmatrix))
    ax = plt.plot(df[0],df[1], 'bo')


import numpy as np
import numba



################### numeric data #########################
@numba.njit(fastmath=True)
def euclidean(x, y):
    """Standard euclidean distance. l2 distance
    ..math::
        D(x, y) = \sqrt{\sum_i (x_i - y_i)^2}
    """
    result = 0.0
    for i in range(x.shape[0]):
        result += (x[i] - y[i]) ** 2
    return np.sqrt(result)

@numba.njit(fastmath=True)
def sqeuclidean(x, y):
    """Standard euclidean distance. l2 distance
    ..math::
        D(x, y) = \sqrt{\sum_i (x_i - y_i)^2}
    """
    result = 0.0
    for i in range(x.shape[0]):
        result += (x[i] - y[i]) ** 2
    return result



@numba.njit()
def manhattan(x, y):
    """Manhatten, taxicab, or l1 distance.
    ..math::
        D(x, y) = \sum_i |x_i - y_i|
    """
    result = 0.0
    for i in range(x.shape[0]):
        result += np.abs(x[i] - y[i])

    return result


@numba.njit()
def canberra(x, y):
    result = 0.0
    for i in range(x.shape[0]):
        denominator = np.abs(x[i]) + np.abs(y[i])
        if denominator > 0:
            result += np.abs(x[i] - y[i]) / denominator

    return result



@numba.njit()
def chebyshev(x, y):
    """Chebyshev or l-infinity distance.
    ..math::
        D(x, y) = \max_i |x_i - y_i|
    """
    result = 0.0
    for i in range(x.shape[0]):
        result = max(result, np.abs(x[i] - y[i]))

    return result



############### binary data ################
@numba.njit()
def jaccard(x, y):
    num_non_zero = 0.0
    num_equal = 0.0
    for i in range(x.shape[0]):
        x_true = x[i] != 0
        y_true = y[i] != 0
        num_non_zero += x_true or y_true
        num_equal += x_true and y_true

    if num_non_zero == 0.0:
        return 0.0
    else:
        return float(num_non_zero - num_equal) / num_non_zero
    
@numba.njit()
def rogers_tanimoto(x, y):
    num_not_equal = 0.0
    for i in range(x.shape[0]):
        x_true = x[i] != 0
        y_true = y[i] != 0
        num_not_equal += x_true != y_true

    return (2.0 * num_not_equal) / (x.shape[0] + num_not_equal)



@numba.njit()
def hamming(x, y):
    result = 0.0
    for i in range(x.shape[0]):
        if x[i] != y[i]:
            result += 1.0

    return float(result) / x.shape[0]


@numba.njit()
def dice(x, y):
    num_true_true = 0.0
    num_not_equal = 0.0
    for i in range(x.shape[0]):
        x_true = x[i] != 0
        y_true = y[i] != 0
        num_true_true += x_true and y_true
        num_not_equal += x_true != y_true

    if num_not_equal == 0.0:
        return 0.0
    else:
        return num_not_equal / (2.0 * num_true_true + num_not_equal)


@numba.njit()
def kulsinski(x, y):
    num_true_true = 0.0
    num_not_equal = 0.0
    for i in range(x.shape[0]):
        x_true = x[i] != 0
        y_true = y[i] != 0
        num_true_true += x_true and y_true
        num_not_equal += x_true != y_true

    if num_not_equal == 0:
        return 0.0
    else:
        return float(num_not_equal - num_true_true + x.shape[0]) / (
            num_not_equal + x.shape[0]
        )
    
@numba.njit()
def sokal_sneath(x, y):
    num_true_true = 0.0
    num_not_equal = 0.0
    for i in range(x.shape[0]):
        x_true = x[i] != 0
        y_true = y[i] != 0
        num_true_true += x_true and y_true
        num_not_equal += x_true != y_true

    if num_not_equal == 0.0:
        return 0.0
    else:
        return num_not_equal / (0.5 * num_true_true + num_not_equal)

    

    
################### both #############
@numba.njit()
def bray_curtis(x, y):
    numerator = 0.0
    denominator = 0.0
    for i in range(x.shape[0]):
        numerator += np.abs(x[i] - y[i])
        denominator += np.abs(x[i] + y[i])

    if denominator > 0.0:
        return float(numerator) / denominator
    else:
        return 0.0


@numba.njit()
def cosine(x, y):
    result = 0.0
    norm_x = 0.0
    norm_y = 0.0
    for i in range(x.shape[0]):
        result += x[i] * y[i]
        norm_x += x[i] ** 2
        norm_y += y[i] ** 2

    if norm_x == 0.0 and norm_y == 0.0:
        return 0.0
    elif norm_x == 0.0 or norm_y == 0.0:
        return 1.0
    else:
        return 1.0 - (result / np.sqrt(norm_x * norm_y))


@numba.njit()
def correlation(x, y):
    mu_x = 0.0
    mu_y = 0.0
    norm_x = 0.0
    norm_y = 0.0
    dot_product = 0.0

    for i in range(x.shape[0]):
        mu_x += x[i]
        mu_y += y[i]

    mu_x /= x.shape[0]
    mu_y /= x.shape[0]

    for i in range(x.shape[0]):
        shifted_x = x[i] - mu_x
        shifted_y = y[i] - mu_y
        norm_x += shifted_x ** 2
        norm_y += shifted_y ** 2
        dot_product += shifted_x * shifted_y

    if norm_x == 0.0 and norm_y == 0.0:
        return 0.0
    elif dot_product == 0.0:
        return 1.0
    else:
        return 1.0 - (dot_product / np.sqrt(norm_x * norm_y))
    
    
    
descriptors_dist = [(euclidean,'euclidean'),
                    (sqeuclidean,'sqeuclidean'),
                    (manhattan,'manhattan'),
                    (canberra,'canberra'),
                    (chebyshev,'chebyshev'),
                    (cosine,'cosine'),
                    (correlation,'correlation'),
                    (bray_curtis,'braycurtis')]



fingerprint_dist = [(jaccard, 'jaccard'),
                    (rogers_tanimoto, 'rogers_tanimoto'),
                    (hamming,'hamming'),
                    (dice, 'dice'),
                    (kulsinski, 'kulsinski'),
                    (sokal_sneath,'sokal_sneath'),
                    (cosine,'cosine'),
                    (correlation,'correlation'),
                    (bray_curtis,'braycurtis')]




def GenNamedDist(descriptors_dist, fingerprint_dist):
    _dist_fuc = {}
    _all = descriptors_dist.copy()
    _all.extend(fingerprint_dist)
    for i in _all:
        _dist_fuc[i[1]] = i[0]
    return _dist_fuc


named_distances = GenNamedDist(descriptors_dist, fingerprint_dist)




if __name__ == '__main__':
    
    import pandas as pd
    
    x = np.random.random_sample(size=(100,2))
    x1 = x.round()
    
    res = {}
    for f,k in descriptors_dist:
        ks = 'descriptors-' + k
        res.update({ks:f(x[:,0], x[:,1])})
        
    for f,k in fingerprint_dist:
        ks = 'fingerprint-' + k
        res.update({ks :f(x1[:,0], x1[:,1])})   
        
    
    print(pd.Series(res))



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 27 14:06:17 2021

@author: Shen Wanxiang
"""

import csv
from collections import defaultdict
from pprint import pprint
import pandas as pd
from scipy.cluster.hierarchy import dendrogram, linkage, to_tree
from scipy.spatial.distance import squareform

itol_header = '''TREE_COLORS
SEPARATOR TAB

#First 3 fields define the node id, type and color
#Possible types are:
#'range': defines a colored range (colored background for labels/clade)
#'clade': defines color/style for all branches in a clade
#'branch': defines color/style for a single branch
#'label': defines font color/style for the leaf label
#'label_background': defines the leaf label background color

#The following additional fields are required:
#for 'range', field 4 defines the colored range label (used in the legend)

#The following additional fields are optional:
#for 'label', field 4 defines the font style ('normal',''bold', 'italic' or 'bold-italic') and field 5 defines the numeric scale factor for the font size (eg. with value 2, font size for that label will be 2x the standard size)
#for 'clade' and 'branch', field 4 defines the branch style ('normal' or 'dashed') and field 5 defines the branch width scale factor (eg. with value 0.5, branch width for that clade will be 0.5 the standard width)

DATA
#NODE_ID TYPE COLOR LABEL_OR_STYLE SIZE_FACTOR
'''

def _getNewick(node, newick, parentdist, leaf_names):
    if node.is_leaf():
        return "%s:%.2f%s" % (leaf_names[node.id], parentdist - node.dist, newick)
    else:
        if len(newick) > 0:
            newick = "):%.2f%s" % (parentdist - node.dist, newick)
        else:
            newick = ");"
        newick = _getNewick(node.get_left(), newick, node.dist, leaf_names)
        newick = _getNewick(node.get_right(), ",%s" % (newick), node.dist, leaf_names)
        newick = "(%s" % (newick)
        return newick
    
def mp2newick(mp, treefile = 'phenotype_tree', leaf_names = None):
    
    df = mp.df_embedding[['colors','Subtypes']]
    
    if leaf_names == None:
        leaf_names = mp.flist #use final list only
    else:
        assert len(leaf_names) == len(df), 'leaf names should be a list with a length of %s' % len(df)
        df.index = leaf_names
        
    linkage_matrix = mp.Z
    tree = to_tree(linkage_matrix, rd=False)
    newick = _getNewick(tree, "", tree.dist, leaf_names = leaf_names)
    
    # write newick file for itol
    with open(treefile + '.nwk', 'w') as f:
        f.write(newick)

    # write dataset file for itol
    df['TYPE'] = 'clade'
    df['STYLE'] = 'normal'
    df = df[['TYPE', 'colors', 'STYLE']]
    with open(treefile + '.txt', 'w') as f:
        f.write(itol_header)
    df.to_csv(treefile + '.txt', mode = 'a', header=None, sep='\t')
    return df
    
def tree(): 
    return defaultdict(tree)

def tree_add(t, path):
    for node in path:
        t = t[node]

def pprint_tree(tree_instance):
    def dicts(t): return {k: dicts(t[k]) for k in t}
    pprint(dicts(tree_instance))

def dfs_to_tree(dfs):
    t = tree()
    for i in range(len(dfs)):
        row = dfs.iloc[i].dropna().tolist()
        tree_add(t, row)
    return t

def tree_to_newick(root):
    items = []
    for k in root.keys():
        s = ''
        if len(root[k].keys()) > 0:
            sub_tree = tree_to_newick(root[k])
            if sub_tree != '':
                s += '(' + sub_tree + ')'
        s += k
        items.append(s)
    return ','.join(items)

def dfs_to_weightless_newick(dfs):
    t = dfs_to_tree(dfs)
    newick_tree = tree_to_newick(t)
    return newick_tree


if __name__ == '__main__':
    
    species_list = pd.read_csv('./species.list.csv', header=None,index_col=0)[1].to_list()
    dfs = pd.Series(species_list).apply(lambda x: dict([i.split('__') for i in x.split('|')])).apply(pd.Series)
    level_dict = {'k':'kingdom', 'p':'phylum', 'c':'class' ,'o':'order' ,'f':'family' ,'g': 'genus','s': 'species'}
    dfs = dfs.rename(columns=level_dict)
    nwk_string = dfs_to_weightless_newick(dfs)
    with open("1.nwk", "w") as f:
        f.write(nwk_string)
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov 21 12:52:49 2018

@author: shenwanxiang

Multi process Run
"""

import time
import pandas as pd
from tqdm import tqdm

from concurrent.futures import ProcessPoolExecutor, wait, as_completed
from multiprocessing import Pool,cpu_count,current_process 
import subprocess


from aggmap.utils.logtools import print_info, print_error, pbar,print_warn


def RunCmd(cmd):
    '''
    input:
        cmd: str
    output:
        status: int, 0 for success
        stdout: str
        stderr: str
        
    '''
    print_info('run command : %s' % cmd)
    
    def swap_log(swap, error = True):
        sinfo = []
        for l in swap.split('\n'):
            if l == '':
                continue
            sinfo.append(l)
        for o in sinfo:
            if error:
                print_error(o) 
            else:
                print_info(o) 
        return            
    output = subprocess.run(cmd, 
                            shell=True, 
                            stdout=subprocess.PIPE, 
                            stderr=subprocess.PIPE, 
                            universal_newlines=True)    
    status = output.returncode
    stdout = output.stdout
    stderr = output.stderr
    
    if status != 0:
        if output.stdout:
             swap_log(output.stdout, error=True)
        if output.stderr:
             swap_log(output.stderr, error=True)
    else:
        if output.stdout:
            swap_log(output.stdout, error=False)
    #return status

    return status, stdout, stderr



def ImapUnorder(processor, iterator, max_workers=10, fail_in_file = './filed.lst'):
    '''
    processor: fuction
    iterator: list or iterator,each element should be a tuple or dict, so that data can be used as ordered 
    '''
    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        
        with open(fail_in_file, 'w+') as f:
            futures = {executor.submit(processor, IdPlusSmile):IdPlusSmile for IdPlusSmile in iterator}
            success, _ = wait(futures)
            with pbar(total = len(futures)) as pb:
                for i in success:
                    IdPlusSmile = futures[i]
                    print_info('deal '+ str(IdPlusSmile))
                    try:
                        data_dict = i.result()
                        yield data_dict
                    except Exception as exc:
                        print_warn('because of the process is dead, input: %s is fialed when deal with %s: %s, so we will deal it automatically' % (IdPlusSmile, processor, exc))
                        
                        try: 
                            yield processor(IdPlusSmile)
                        except:
                            f.write(str(IdPlusSmile)+'\n')
                            print_error(' input: %s is fialed when deal with %s: %s' % (IdPlusSmile, processor, exc))
                    pb.update(1)

                    


                    
def MultiProcessUnorderedBarRun(func, deal_list, n_cpus=None):
    if n_cpus ==None:
        N_CPUS = cpu_count()
    else:
        N_CPUS = int(n_cpus)
    print_info('the number of process is %s' % N_CPUS)
    
    p = Pool(N_CPUS)
    res_list = []
    with pbar(total = len(deal_list), ascii=True) as pb:
        for res in p.imap_unordered(func, deal_list):
            pb.update(1)
            res_list.append(res)
    p.close()
    p.join()
    return res_list



def MultiProcessRun(func, deal_list, n_cpus=None):
    
    '''
    input:
        func: function to do with each element in the deal_list
        deal_list: list to be done
        n_cpus: use the number of cpus
    output:
        list of the return result for each func
    '''
    
    #round_c = [deal_list[i:i+batch_size] for i  in range(0, len(deal_list), batch_size)]
    #mata thinking: https://my.oschina.net/leejun2005/blog/203148
    if n_cpus ==None:
        N_CPUS = cpu_count()
    else:
        N_CPUS = int(n_cpus)

    print_info('the number of process is %s' % N_CPUS)

    pool = Pool(N_CPUS)
    a = pool.map(func, deal_list)
    pool.close()
    pool.join()
    return a






########### ordered map reduce  ##############
def _decorate_func(func, i, j):
    return [i, func(j)]

def _executor(func, series, n_cpus = 4):
    with ProcessPoolExecutor(max_workers=n_cpus) as executor:
        futures = [executor.submit(_decorate_func, func, i, j) for i,j in series.iteritems()]
    return futures


def MultiExecutorRun(func, deal_list, n_cpus = 4, tqdm_args = {'unit':'one'}):
    
    '''
    input:
        func: function to do with each element in the deal_list
        deal_list: list to be done
        n_cpus: use the number of cpus
        tqdm_args: args for tqdm
    output:
        list of the return value for each func
    '''
    lst  =list(deal_list)
    series = pd.Series(lst)
    
    futures = _executor(func, series, n_cpus = n_cpus)
    args = {
        'total': len(deal_list),
        'unit': 'one',
        'ascii': True,
        'unit_scale': True,
        'leave': True
    }
    args.update(tqdm_args)
    
    print_info(args)
    
    results = []
    indexs = []
    for f in tqdm(as_completed(futures), **args):
        #print(f)
        idx, result = f.result()
        indexs.append(idx)
        results.append(result)
    
    res = pd.Series(results,index=indexs)
    #sort unordered result
    ordered_lst = res.sort_index().tolist()
    return ordered_lst
    